import numpy as np
from typing import Dict, List, Set, Tuple, Union

from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace_root as PST
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidNumberOfArgumentsError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLInvalidObjectError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLRepeatedVariableError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError

from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.Core.Debug.Logger import Logger
from pyRDDLGym.Core.Parser.expr import Expression


class RDDLTracedObjects:
    '''A generic container for storing traced information for a RDDL file.'''
    
    def __init__(self) -> None:
        self._current_id = 0
        self._current_root = ''
        self._cached_objects_in_scope = []
        self._cached_object_type = []
        self._cached_is_fluent = []
        self._cached_is_fluent_cpf = {}
        self._cached_sim_info = []
        self._expr_from_id = {}
        
    def _append(self, expr, objects, obj_type, is_fluent, info) -> None:
        expr.id = self._current_id
        self._current_id += 1
                
        self._cached_objects_in_scope.append(objects)
        self._cached_object_type.append(obj_type)
        self._cached_is_fluent.append(is_fluent)
        self._cached_sim_info.append(info)
        self._expr_from_id[expr.id] = expr
        
    def cached_objects_in_scope(self, expr: Expression):
        '''Returns the free variables/parameters in the scope of expression.'''
        return self._cached_objects_in_scope[expr.id]
    
    def cached_object_type(self, expr: Expression) -> Union[str, None]:
        '''Returns the returned object type of expression or None if not an object.'''
        return self._cached_object_type[expr.id]
    
    def cached_is_fluent(self, expr: Expression) -> bool:
        '''Returns whether the expression is fluent or non-fluent.'''
        return self._cached_is_fluent[expr.id]
    
    def cached_is_fluent_cpf(self, name: str) -> bool:
        '''Returns whether the CPF given by name is a fluent expression.'''
        return self._cached_is_fluent_cpf[name]
    
    def cached_sim_info(self, expr: Expression) -> object:
        '''Returns compiled info that is specific to the expression.'''
        return self._cached_sim_info[expr.id]
    
    def lookup(self, identifier: int) -> Expression:
        '''Returns the expression with given identifier, or None if does not 
        exist.'''
        return self._expr_from_id.get(identifier, None)
    

def py_enum(**enums):
    return type('Enum', (), enums)

    
class RDDLObjectsTracer:
    '''Performs static/compile-time tracing of a RDDL AST representation and
    annotates nodes with info about objects that appear inside expressions.'''
    
    # for multivariate sampling: how many dimensions sample has
    REQUIRED_DIST_PVARS = {
        'MultivariateNormal': 1,
        'MultivariateStudent': 1,
        'Dirichlet': 1,
        'Multinomial': 1
    }
    
    # for multivariate sampling: how many _ identifiers each argument has
    REQUIRED_DIST_UNDERSCORES = {
        'MultivariateNormal': (1, 2),  # mean, covariance
        'MultivariateStudent': (1, 2, 0),  # mean, covariance, df
        'Dirichlet': (1,),  # weights
        'Multinomial': (0, 1)  # trials, probabilities
    }
    
    # operation codes on pvariable tensors
    NUMPY_OP_CODE = py_enum(
        NESTED_SLICE=-1,  # nested pvariables detected: slice required
        EINSUM=0,  # duplicated variables like fluent(?x, ?x) - reduction required
        TRANSPOSE=1,  # evaluation order differs from outerscope -- reorder required
        NOOP=2  # a scalar pvariable for example does not require any operation
    )
    
    def __init__(self, rddl: PlanningModel, logger: Logger=None,
                 cpf_levels: Dict[int, Set[str]]=None) -> None:
        '''Creates a new objects tracer object for the given RDDL domain.
        
        :param rddl: the RDDL domain to trace
        :param logger: to log compilation information during tracing to file
        :param cpf_levels: pre-computed CPF levels (will be computed internally
        if None)
        '''
        self.rddl = rddl
        self.logger = logger
        self.cpf_levels = cpf_levels
            
    @staticmethod
    def _check_not_object(arg, expr, out, msg):
        obj_type = out.cached_object_type(arg)
        if obj_type is not None:
            raise RDDLTypeError(
                f'{msg} can not be an object of type <{obj_type}>.\n' + 
                PST(expr, out._current_root)) 
            
    def trace(self) -> RDDLTracedObjects:
        '''Traces all expressions in the current RDDL file and annotates
        AST nodes with object information. 
        
        Important notes: 
        (1) tracing annotates intermediate expressions in the AST with unique 
        identifiers; if the shape or structure of the AST changes, it is required
        to call trace() again to re-populate these IDs
        (2) the identifiers are cached internally in each Expression node: care
        must therefore be taken if object references in two different AST point
        to the same underlying Expression object, since they could require 
        different IDs; this generally shouldn't happen in typical use cases, but
        any manual AST construction can simply ensure that all expressions to be
        shared across different AST instances are (deep) copied.'''   
        rddl = self.rddl 
        out = RDDLTracedObjects()   
        
        # compute trace order for CPFs
        levels = self.cpf_levels
        if levels is None:
            levels = RDDLLevelAnalysis(rddl).compute_levels()
            
        # trace CPFs
        for cpfs in levels.values():
            for cpf in cpfs:
                objects, expr = rddl.cpfs[cpf]
                
                # check that the parameters are unique
                pvars = [pvar for (pvar, _) in objects]
                if len(set(pvars)) != len(pvars):
                    raise RDDLRepeatedVariableError(
                        f'Repeated parameter(s) {pvars} in definition of CPF <{cpf}>.')
                
                # trace the expression
                out._current_root = f'CPF {cpf}'
                self._trace(expr, objects, out)
                out._cached_is_fluent_cpf[cpf] = out.cached_is_fluent(expr)

                # for domain-object valued check that type matches expression output
                cpf_range = rddl.variable_ranges[cpf]
                expr_range = out.cached_object_type(expr)
                if (cpf_range in rddl.enums and expr_range != cpf_range) \
                or (cpf_range not in rddl.objects and expr_range is not None):
                    raise RDDLTypeError(
                        f'CPF <{cpf}> expression expects type <{cpf_range}>, '
                        f'got expression of type <{expr_range}>.')

        # trace reward, check not object value
        out._current_root = 'reward'
        self._trace(rddl.reward, [], out)
        RDDLObjectsTracer._check_not_object(rddl.reward, rddl.reward, out, 'reward')
        
        # trace constraints, check not object
        for (i, expr) in enumerate(rddl.invariants):
            out._current_root = f'Invariant {i + 1}'
            self._trace(expr, [], out)
            RDDLObjectsTracer._check_not_object(expr, expr, out, out._current_root)
        for (i, expr) in enumerate(rddl.preconditions):
            out._current_root = f'Precondition {i + 1}'
            self._trace(expr, [], out)
            RDDLObjectsTracer._check_not_object(expr, expr, out, out._current_root)
        for (i, expr) in enumerate(rddl.terminals):
            out._current_root = f'Termination {i + 1}'
            self._trace(expr, [], out)
            RDDLObjectsTracer._check_not_object(expr, expr, out, out._current_root)
        
        # log the fluent types
        if self.logger is not None:
            message = f'[info] computed whether each CPF expression is fluent:\n'
            for cpfs in levels.values():
                for cpf in cpfs:
                    is_fluent = out._cached_is_fluent_cpf[cpf]
                    message += f'\t{cpf}: {is_fluent}\n'
            self.logger.log(message)
            
        return out
        
    # ===========================================================================
    # start of tracing subroutines
    # ===========================================================================
        
    def _trace(self, expr, objects, out):
        etype, _ = expr.etype
        if etype == 'constant':
            self._trace_constant(expr, objects, out)
        elif etype == 'pvar':
            self._trace_pvar(expr, objects, out)
        elif etype == 'arithmetic':
            self._trace_arithmetic(expr, objects, out)
        elif etype == 'relational':
            self._trace_relational(expr, objects, out)
        elif etype == 'boolean':
            self._trace_logical(expr, objects, out)
        elif etype == 'aggregation':
            self._trace_aggregation(expr, objects, out)
        elif etype == 'func':
            self._trace_func(expr, objects, out)
        elif etype == 'control':
            self._trace_control(expr, objects, out)
        elif etype == 'randomvar':
            self._trace_random(expr, objects, out)
        elif etype == 'randomvector':
            self._trace_random_vector(expr, objects, out)
        elif etype == 'matrix':
            self._trace_matrix(expr, objects, out)
        else:
            raise RDDLNotImplementedError(
                f'Internal error: expression type {etype} is not supported.\n' + 
                PST(expr, out._current_root))
    
    # ===========================================================================
    # leaves
    # ===========================================================================
    
    def _trace_constant(self, expr, objects, out):
        if objects:
            shape = self.rddl.object_counts((ptype for (_, ptype) in objects))
            cached_value = np.full(shape=shape, fill_value=expr.args)
        else:
            cached_value = expr.args
            
        out._append(expr, objects, None, False, cached_value)
        
    def _trace_pvar(self, expr, objects, out):
        var, pvars = expr.args   
        rddl = self.rddl
        
        # free variable (e.g., ?x) treated as array
        # first element True indicates value is to be returned directly by sim
        if rddl.is_free_variable(var):
            obj_to_index = {pobj: i for (i, (pobj, _)) in enumerate(objects)}
            index_of_var = obj_to_index.get(var, None)
                         
            # check var is valid in scope "objects"
            if index_of_var is None:
                raise RDDLInvalidObjectError(
                    f'Free variable <{var}> is not defined in outer scope, '
                    f'must be one of {set(obj_to_index.keys())}. '
                    f'Please check expression for <{out._current_root}>.')
            
            # create an array whose shape matches objects
            # along axis equal to index_of_var_in_objects values are (0, 1, ...)
            ptypes = [ptype for (_, ptype) in objects]
            shape = rddl.object_counts(ptypes)            
            cached_value = np.arange(shape[index_of_var])
            cached_value = cached_value[(...,) + (np.newaxis,) * len(ptypes[1:])]
            cached_value = np.swapaxes(cached_value, axis1=0, axis2=index_of_var)
            cached_value = np.broadcast_to(cached_value, shape=shape)
            cached_value = (True, cached_value)
            
            prange = ptypes[index_of_var]
            obj_type = prange if (prange in rddl.enums or prange in rddl.objects) else None
            out._append(expr, objects, obj_type, False, cached_value)
        
        # object can only be defined in domain - map to canonical index
        # first element True indicates value is to be returned directly by sim
        elif not pvars and rddl.is_object(
            var, msg='\n' + PST(expr, out._current_root)): 
            
            # check var is a domain object
            literal = rddl.object_name(var)            
            enum_type = rddl.objects_rev[literal]  
            if enum_type not in rddl.enums:
                raise RDDLInvalidObjectError(
                    f'Object <{var}> must be of a domain-defined object type, '
                    f'got type <{enum_type}>. '
                    f'Please check expression for <{out._current_root}>.')
            
            # map to canonical index - for pvariable fill an array with it
            const = rddl.index_of_object[literal]
            if objects:
                shape = rddl.object_counts((ptype for (_, ptype) in objects))
                cached_value = np.full(shape=shape, fill_value=const)
            else:
                cached_value = const            
            cached_value = (True, cached_value)
            
            out._append(expr, objects, enum_type, False, cached_value)
        
        # if the pvar has free variables (e.g., ?x)...
        else:
            
            # check if pvar is fluent
            primed_var = rddl.next_state.get(var, var)
            if primed_var in rddl.cpfs:
                is_fluent = out._cached_is_fluent_cpf.get(primed_var, None)
                if is_fluent is None:
                    is_fluent = True
                    if self.logger is not None:
                        message = (
                            f'[warning] cannot establish whether CPF is fluent:'
                            f'\n\taddress of expression ={str(expr)}'
                            f'\n\tCPF to check          ={primed_var}\n'
                        )
                        self.logger.log(message)
            else:
                is_fluent = rddl.variable_types[var] != 'non-fluent'

            # recursively trace nested pvariables
            if pvars is not None: 
                for arg in pvars:
                    if isinstance(arg, Expression):
                        self._trace(arg, objects, out)
                        is_fluent = is_fluent or out.cached_is_fluent(arg)
            
            # find a way to map value tensor of expr to match objects
            cached_sim_info = (False, self._map(expr, objects, out))    
               
            prange = rddl.variable_ranges.get(var, None)
            obj_type = prange if (prange in rddl.enums or prange in rddl.objects) else None
            out._append(expr, objects, obj_type, is_fluent, cached_sim_info)
        
    def _map(self, expr: Expression,
             objects: Union[List[Tuple[str, str]], None],
             out: RDDLTracedObjects) -> Tuple[object, ...]:
        '''Returns information needed to reshape and transform value tensor for a
        parameterized variable to match desired output signature.
        
        :param expr: outputs of expression value tensor to transform
        :param objects: a list of tuples (objecti, typei) representing the
            desired signature of the output value tensor
        :param cache where child of expr properties have been stored
        '''
        var, args = expr.args
        rddl = self.rddl
        
        # check that the number of input objects match fluent type definition
        if args is None:
            args = []
        args_types = rddl.param_types.get(var, [])
        if len(args) != len(args_types):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Variable <{var}> requires {len(args_types)} argument(s), '
                f'got {len(args)}.\n' + PST(expr, out._current_root))
        
        # for vector distributions, occurrences of _ are mapped to unique new 
        # objects ?_1, ?_2, ... because we have to take care not to einsum them;
        # they will correspond to the last axes of the resulting value array
        # e.g., values(?p, ?q, ... ?_1, ?_2, ...);
        # this way ?p, ?q... can be interpreted as "batch" dimensions and 
        # ?_1, ?_2, ... as the "image" dimensions in an equivalent ML problem;
        # this is the convention of most numpy and JAX batched subroutines anyway
        new_objects, underscore_names = [], {}
        for (i, arg) in enumerate(args):
            if arg == '_':
                new_pvar = f'?_{1 + len(new_objects)}'
                new_objects.append((new_pvar, args_types[i]))
                underscore_names[i] = new_pvar
        objects = objects + new_objects
        
        # test for nested expressions
        is_arg_expr = [isinstance(arg, Expression) for arg in args]
        nested = np.any(is_arg_expr)
                
        # literals are converted to canonical indices in their object list
        # and used to extract from the value tensor at the corresponding axis
        # 1. if there are nested variables, then they are left as None slices
        #    since their values are only known at run time
        # 2. if there are free variables ?x among nested variables, then they
        #    are reshaped to match objects
        object_shape = rddl.object_counts((ptype for (_, ptype) in objects))
        object_index = {obj: i for (i, (obj, _)) in enumerate(objects)}
        slices = [slice(None)] * len(args)
        do_slice = False
        permuted = []
        for (i, arg) in enumerate(args):
            
            # is a nested fluent (e.g., fluent(another-fluent(?x)) )
            if is_arg_expr[i]:
                
                # check that type of the inner fluent matches what var expects
                obj_type = out.cached_object_type(arg)
                if args_types[i] != obj_type: 
                    if obj_type is None:
                        obj_type = 'real, int or bool'
                    raise RDDLTypeError(
                        f'Argument {i + 1} of variable <{var}> expects object '
                        f'of type <{args_types[i]}>, got nested expression '
                        f'of type <{obj_type}>.\n' + PST(expr, out._current_root))
                
                # leave slice blank since it's filled at runtime
                slices[i] = None
                
            # is an object literal (e.g., @x)
            elif rddl.is_object(arg, msg='\n' + PST(expr, out._current_root)):
                
                # check that the type of the object is correct
                literal = rddl.object_name(arg)
                enum_type = rddl.objects_rev[literal]
                if args_types[i] != enum_type: 
                    raise RDDLTypeError(
                        f'Argument {i + 1} of variable <{var}> expects object '
                        f'of type <{args_types[i]}>, got <{arg}> '
                        f'of type <{enum_type}>.\n' + PST(expr, out._current_root))
                
                # extract value at current dimension at object's canonical index
                slices[i] = rddl.index_of_object[literal]
                do_slice = True
            
            # is a free object (e.g., ?x)
            else:
                
                # a sampling dimension is mapped to its unique quantifier ?_i
                arg = underscore_names.get(i, arg)
                
                # make sure argument is well defined
                index_of_arg = object_index.get(arg, None)
                if index_of_arg is None:
                    raise RDDLInvalidObjectError(
                        f'Undefined argument <{arg}> at position {i + 1} '
                        f'of variable <{var}>.\n' + PST(expr, out._current_root))
                
                # make sure type of argument is correct
                _, ptype = objects[index_of_arg]
                if ptype != args_types[i]:
                    raise RDDLTypeError(
                        f'Argument {i + 1} of variable <{var}> expects object '
                        f'of type <{args_types[i]}>, got <{arg}> '
                        f'of type <{ptype}>.\n' + PST(expr, out._current_root))
                                      
                # if nesting is found, then free variables like ?x are implicitly 
                # converted to arrays with shape of objects
                # this way, the slice value array has shape that matches objects
                if nested:
                    indices = np.arange(len(rddl.objects[ptype]))
                    newshape = [1] * len(objects)
                    newshape[index_of_arg] = indices.size
                    indices = np.reshape(indices, newshape=newshape)
                    indices = np.broadcast_to(indices, shape=object_shape)
                    slices[i] = indices 
                
                # if no nesting, then we use einsum or transpose operations
                else:
                    permuted.append(index_of_arg)
                                                       
        slices = tuple(slices) if do_slice or nested else ()
        len_after_slice = len(permuted)
        
        # compute the mapping function as follows:
        # 0. first assume all literals are "sliced out" of the value tensor
        #    if there is nesting then a slice is sufficient and do nothing else
        # 1. append new axes to value tensor equal to # of missing variables
        # 2. broadcast new axes to the desired shape (# of objects of each type)
        # 3. rearrange the axes as needed to match the desired variables in order
        #    3a. in most cases, it suffices to use np.transform (cheaper)
        #    3b. in cases where we have a more complex contraction like 
        #        fluent(?x) = matrix(?x, ?x), we will use np.einsum
        if nested:
            new_axis = None
            new_shape = None
            op_args = None
            op_code = RDDLObjectsTracer.NUMPY_OP_CODE.NESTED_SLICE
        else:
            
            # update permutation based on objects not in args
            covered = set(permuted)
            permuted += [i for i in range(len(objects)) if i not in covered]
            
            # store the arguments for each operation: 
            # 0 means einsum, 1 means transpose, and others (-1, 2) are no-op
            new_axis = tuple(range(len_after_slice, len(permuted)))  
            new_shape = tuple(object_shape[i] for i in permuted)
            objects_range = list(range(len(objects)))        
            if len(covered) != len_after_slice:
                op_args = (permuted, objects_range)
                op_code = RDDLObjectsTracer.NUMPY_OP_CODE.EINSUM
            elif permuted != objects_range:
                op_args = tuple(np.argsort(permuted))  # inverse permutation
                op_code = RDDLObjectsTracer.NUMPY_OP_CODE.TRANSPOSE
            else:
                op_args = None
                op_code = RDDLObjectsTracer.NUMPY_OP_CODE.NOOP
        
        # log information about the new transformation
        if self.logger is not None:
            message = (
                f'[info] computing info for pvariable tensor transformation:' 
                f'\n\taddress of expression   ={super(Expression, expr).__str__()}'
                f'\n\tvariable                ={var}'
                f'\n\tvariable evaluated at   ={list(zip(args, args_types))}'
                f'\n\tfree object(s) in scope ={objects}'
                f'\n\tslice                   ={slices}'
                f'\n\tnew axes to append      ={new_axis}'
                f'\n\tbroadcast shape         ={new_shape}'
                f'\n\ttransform operation     ={op_code} with argument(s) {op_args}\n'
            )
            self.logger.log(message)
            
        return (slices, new_axis, new_shape, op_code, op_args)
    
    # ===========================================================================
    # compound expressions
    # ===========================================================================
    
    def _trace_arithmetic(self, expr, objects, out): 
        _, op = expr.etype
        is_fluent = False
        for (i, arg) in enumerate(expr.args):
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)
        
            # cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Argument {i + 1} of operator {op}')
              
        out._append(expr, objects, None, is_fluent, None)
        
    def _trace_relational(self, expr, objects, out):
        _, op = expr.etype
        args = expr.args
        is_fluent = False
        for arg in args:
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)         
        
        # can not mix different object types or primitive and object types
        obj_types = set(map(out.cached_object_type, args))
        if len(obj_types) != 1:
            raise RDDLTypeError(
                f'Relational operator {op} can not compare arguments '
                f'of different object types {obj_types}.\n' + 
                PST(expr, out._current_root))
        
        # can not use operator besides == and ~= to compare object types
        obj_type, = obj_types
        if obj_type is not None and op != '==' and op != '~=':
            raise RDDLTypeError(
                f'Relational operator {op} is not valid '
                f'for comparing objects of type {obj_type}.\n' + 
                PST(expr, out._current_root))
        
        out._append(expr, objects, None, is_fluent, None)
    
    def _trace_logical(self, expr, objects, out):
        is_fluent = False
        for (i, arg) in enumerate(expr.args):
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)
            
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Argument {i + 1} of operator {expr.etype[1]}')
        
        out._append(expr, objects, None, is_fluent, None)
    
    def _trace_func(self, expr, objects, out):
        is_fluent = False
        for (i, arg) in enumerate(expr.args):
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)
            
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Argument {i + 1} of function {expr.etype[1]}') 
        
        out._append(expr, objects, None, is_fluent, None)
            
    # ===========================================================================
    # aggregation
    # ===========================================================================
    
    def _check_iteration_variables(self, objects, iter_objects, expr, out):
        
        # check for undefined types
        rddl = self.rddl
        bad_types = {ptype 
                     for (_, ptype) in iter_objects 
                     if ptype not in rddl.objects}
        if bad_types:
            raise RDDLTypeError(
                f'Type(s) {bad_types} are not defined, '
                f'must be one of {set(rddl.objects.keys())}.\n' + 
                PST(expr, out._current_root))
        
        # check for valid type arguments
        scope_vars = {var for (var, _) in objects}
        seen_vars = set()
        for (var, _) in iter_objects:
            
            # check that there is no duplicated iteration variable
            if var in seen_vars:
                raise RDDLRepeatedVariableError(
                    f'Iteration variable <{var}> is repeated.\n' + 
                    PST(expr, out._current_root))             
            seen_vars.add(var)
             
            # check if iteration variable is same as one defined in outer scope
            # since there is ambiguity to which is referred I raise an error
            if var in scope_vars:
                raise RDDLRepeatedVariableError(
                    f'Iteration variable <{var}> is already defined '
                    f'in outer scope.\n' + PST(expr, out._current_root))
        
    def _trace_aggregation(self, expr, objects, out):
        _, op = expr.etype
        * pvars, arg = expr.args
        
        # cache and read reduced axes tensor info for the aggregation
        # axes of new free variables in aggregation are appended to value array
        iter_objects = [ptype for (_, ptype) in pvars]
        self._check_iteration_variables(objects, iter_objects, expr, out)
        new_objects = objects + iter_objects
        reduced_axes = tuple(range(len(objects), len(new_objects)))   
        cached_sim_info = (new_objects, reduced_axes)
        
        # argmax/argmin require exactly one parameter argument
        enum_type = None
        if op == 'argmin' or op == 'argmax':
            if len(pvars) != 1:
                raise RDDLInvalidNumberOfArgumentsError(
                    f'Aggregation <{op}> requires one iteration variable, '
                    f'got {len(pvars)}.\n' + PST(expr, out._current_root))
            cached_sim_info = (new_objects, reduced_axes[0])
            (_, (_, enum_type)), = pvars
        
        # trace the aggregated expression with the new objects
        self._trace(arg, new_objects, out)
        is_fluent = out.cached_is_fluent(arg)
        
        # argument cannot be object type
        RDDLObjectsTracer._check_not_object(
            arg, expr, out, f'Argument of aggregation {op}')     
        
        out._append(expr, objects, enum_type, is_fluent, cached_sim_info)
        
        # log information about aggregation operation
        if self.logger is not None:
            message = (f'[info] computing object info for aggregation:'
                       f'\n\taggregation variables(s)      ={pvars}'
                       f'\n\tfree object(s) in outer scope ={objects}'
                       f'\n\tfree object(s) in inner scope ={new_objects}'
                       f'\n\taggregation operation         ={op}'
                       f'\n\taggregation axes              ={reduced_axes}\n')
            self.logger.log(message)        
        
    # ===========================================================================
    # control flow
    # ===========================================================================
    
    def _trace_control(self, expr, objects, out):
        _, op = expr.etype
        if op == 'if':
            self._trace_if(expr, objects, out)
        elif op == 'switch':
            self._trace_switch(expr, objects, out)
            
    def _trace_if(self, expr, objects, out):
        pred, *branches = expr.args
        
        # trace predicate
        self._trace(pred, objects, out)
        RDDLObjectsTracer._check_not_object(
            pred, expr, out, f'Predicate of if statement')  
        
        # trace branches
        is_fluent = out.cached_is_fluent(pred)
        for arg in branches:
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)
        
        # can not mix different object types or primitive and object types
        obj_types = set(map(out.cached_object_type, branches))
        if len(obj_types) != 1:
            raise RDDLTypeError(
                f'Different branches in if statement cannot produce values '
                f'of different object types {obj_types}.\n' + 
                PST(expr, out._current_root))     
    
        obj_type, = obj_types
        out._append(expr, objects, obj_type, is_fluent, None)
        
    def _trace_switch(self, expr, objects, out):
        rddl = self.rddl
        pred, *cases = expr.args
        
        # predicate must be a pvar
        if not pred.is_pvariable_expression():
            raise RDDLNotImplementedError(
                f'Switch predicate is not a pvariable.\n' + 
                PST(expr, out._current_root))
            
        # type in pvariables scope must be a domain object
        var, _ = pred.args
        enum_type = rddl.variable_ranges.get(var, None)
        if enum_type not in rddl.enums:
            raise RDDLTypeError(
                f'Type <{enum_type}> of switch predicate <{var}> is not a '
                f'domain-defined object, must be one of {rddl.enums}.\n' + 
                PST(expr, out._current_root))
            
        # default statement becomes ("default", expr)
        case_dict = dict(
            (case_value if case_type == 'case' else (case_type, case_value)) 
            for (case_type, case_value) in cases
        )    
        
        # strip @ from any cases
        case_dict = {rddl.object_name(key): value 
                     for (key, value) in case_dict.items()}
        
        # check for duplicated cases
        if len(case_dict) != len(cases):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Duplicated literal or default case(s).\n' + 
                PST(expr, out._current_root))
        
        # order cases by canonical ordering of objects
        cached_sim_info = self._order_cases(enum_type, case_dict, expr, out)
        
        # trace predicate and cases
        self._trace(pred, objects, out)
        is_fluent = out.cached_is_fluent(pred)
        for arg in case_dict.values():
            self._trace(arg, objects, out)
            is_fluent = is_fluent or out.cached_is_fluent(arg)
        
        # can not mix different object types or primitive and object types
        obj_types = set(map(out.cached_object_type, case_dict.values()))
        if len(obj_types) != 1:
            raise RDDLTypeError(
                f'Case expressions in switch statement cannot produce values '
                f'of different object types {obj_types}\n' + 
                PST(expr, out._current_root))    
                                
        obj_type, = obj_types
        out._append(expr, objects, obj_type, is_fluent, cached_sim_info)
        
    def _order_cases(self, enum_type, case_dict, expr, out): 
        rddl = self.rddl
        enum_values = rddl.objects[enum_type]
        
        # check that all literals belong to enum_type
        for _case in case_dict:
            if _case != 'default' and rddl.objects_rev.get(_case, None) != enum_type:
                raise RDDLInvalidObjectError(
                    f'Object <{_case}> does not belong to type <{enum_type}>, '
                    f'must be one of {set(enum_values)}.\n' + 
                    PST(expr, out._current_root))
        
        # store expressions in order of canonical literal index
        expressions = [None] * len(enum_values)
        for obj in enum_values:
            arg = case_dict.get(obj, None)
            if arg is not None: 
                index = rddl.index_of_object[obj]
                expressions[index] = arg
        
        # if default statement is missing, cases must be comprehensive
        default_expr = case_dict.get('default', None)
        if default_expr is None:
            for (i, arg) in enumerate(expressions):
                if arg is None:
                    raise RDDLUndefinedVariableError(
                        f'Object <{enum_values[i]}> of type <{enum_type}> '
                        f'is missing in case list.\n' + PST(expr, out._current_root))
        
        # log cases ordering
        if self.logger is not None:
            active_expr = [i for (i, e) in enumerate(expressions) if e is not None]
            message = (f'[info] computing case info for {expr.etype[1]}:'
                       f'\n\tenum type ={enum_type}'
                       f'\n\tcases     ={active_expr}'
                       f'\n\tdefault   ={default_expr is not None}\n')
            self.logger.log(message)     
        
        return (expressions, default_expr)
    
    # ===========================================================================
    # random variable
    # ===========================================================================
    
    def _trace_random(self, expr, objects, out):
        _, name = expr.etype
        if name == 'Discrete' or name == 'UnnormDiscrete':
            self._trace_discrete(expr, objects, out)
        elif name == 'Discrete(p)' or name == 'UnnormDiscrete(p)':
            self._trace_discrete_pvar(expr, objects, out)
        else:
            self._trace_random_other(expr, objects, out)
                
    def _trace_discrete(self, expr, objects, out):
        rddl = self.rddl
        (_, enum_type), *cases = expr.args
            
        # object type must be a valid domain object type
        if enum_type not in rddl.enums:
            raise RDDLTypeError(
                f'Type <{enum_type}> in Discrete distribution is not a '
                f'domain-defined object, must be one of {rddl.enums}.\n' + 
                PST(expr, out._current_root))        
        case_dict = dict(case_tup for (_, case_tup) in cases) 
        
        # strip @ from any cases       
        case_dict = {rddl.object_name(key): value 
                     for (key, value) in case_dict.items()}
        
        # no duplicate cases are allowed
        if len(case_dict) != len(cases):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Duplicated literal or default case(s).\n' + 
                PST(expr, out._current_root))
        
        # no default cases are allowed
        if 'default' in case_dict:
            raise RDDLNotImplementedError(
                f'Default case not allowed in Discrete distribution.\n' + 
                PST(expr, out._current_root))
            
        # order enum cases by canonical ordering of literals
        cached_sim_info, _ = self._order_cases(enum_type, case_dict, expr, out)
    
        # trace each case expression
        for (i, arg) in enumerate(case_dict.values()):
            self._trace(arg, objects, out)
            
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Expression in case {i + 1} of Discrete') 
        
        out._append(expr, objects, enum_type, True, cached_sim_info)
    
    def _trace_discrete_pvar(self, expr, objects, out):
        * pvars, args = expr.args
        
        # check number of iteration variables and arguments
        if len(pvars) != 1:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Discrete requires one iteration variable, got {len(pvars)}.\n' + 
                PST(expr, out._current_root))
        elif len(args) != 1:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Discrete requires one argument, got {len(args)}.\n' + 
                PST(expr, out._current_root))
        
        # sampling variables are appended to scope free variables        
        iter_objects = [pvar for (_, pvar) in pvars]     
        self._check_iteration_variables(objects, iter_objects, expr, out)
        new_objects = objects + iter_objects
        
        # trace the arguments
        for (i, arg) in enumerate(args):
            self._trace(arg, new_objects, out)
                
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Expression in case {i + 1} of Discrete') 

        (_, (_, enum_type)), = pvars
        out._append(expr, objects, enum_type, True, None)

    def _trace_random_other(self, expr, objects, out):
        for (i, arg) in enumerate(expr.args):
            self._trace(arg, objects, out)
                
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Argument {i + 1} of {expr.etype[1]}') 
        
        out._append(expr, objects, None, True, None)
        
    # ===========================================================================
    # random vector
    # ===========================================================================
    
    def _trace_random_vector(self, expr, objects, out):
        _, op = expr.etype
        sample_pvars, args = expr.args
        
        # determine how many instances of _ should appear in each argument
        required_sample_pvars = RDDLObjectsTracer.REQUIRED_DIST_PVARS.get(op, None)
        required_underscores = RDDLObjectsTracer.REQUIRED_DIST_UNDERSCORES.get(op, None)
        if required_sample_pvars is None or required_underscores is None:
            raise RDDLNotImplementedError(
                f'Internal error: distribution {op} is not supported.\n' + 
                PST(expr, out._current_root))
        
        # check the number of sample parameters is correct
        if len(sample_pvars) != required_sample_pvars:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Distribution {op} requires {required_sample_pvars} sampling '
                f'parameter(s), got {len(sample_pvars)}.\n' + 
                PST(expr, out._current_root))
        
        # check that all sample_pvars are defined in the outer scope
        scope_pvars = {pvar: i for (i, (pvar, _)) in enumerate(objects)}
        bad_pvars = {pvar for pvar in sample_pvars if pvar not in scope_pvars}
        if bad_pvars:
            raise RDDLInvalidObjectError(
                f'Sampling parameter(s) {bad_pvars} of {op} are not defined in '
                f'outer scope, must be one of {set(scope_pvars.keys())}.\n' + 
                PST(expr, out._current_root))
        
        # check duplicates in sample_pvars
        sample_pvar_set = set(sample_pvars)
        if len(sample_pvar_set) != len(sample_pvars):
            raise RDDLInvalidNumberOfArgumentsError(
                f'Sampling parameter(s) {sample_pvars} of {op} are duplicated.\n' + 
                PST(expr, out._current_root))
        
        # sample_pvars are excluded when tracing arguments (e.g., mean)
        # because they are only introduced through sampling which follows after 
        # evaluation of the arguments in a depth-first traversal
        batch_objects = [pvar for pvar in objects if pvar[0] not in sample_pvar_set]
         
        # trace all parameters
        enum_types = set()
        for (i, arg) in enumerate(args):
            
            # sample_pvars cannot be argument parameters
            name, pvars = arg.args
            if pvars is None:
                pvars = []
            bad_pvars = {pvar for pvar in pvars if pvar in sample_pvar_set}
            if bad_pvars:
                raise RDDLInvalidObjectError(
                    f'Parameter(s) {bad_pvars} of argument <{name}> at position '
                    f'{i + 1} of {op} can not be sampling parameter(s) '
                    f'{sample_pvar_set}.\n' + PST(expr, out._current_root))
            
            # check number of _ parameters is valid in argument
            underscores = [j for (j, pvar) in enumerate(pvars) if pvar == '_']            
            if len(underscores) != required_underscores[i]:
                raise RDDLInvalidNumberOfArgumentsError(
                    f'Argument <{name}> at position {i + 1} of {op} must contain '
                    f'{required_underscores[i]} sampling parameter(s) _, '
                    f'got {len(underscores)}.\n' + PST(expr, out._current_root))
            
            # trace argument
            self._trace(arg, batch_objects, out)
            
            # argument cannot be object type
            RDDLObjectsTracer._check_not_object(
                arg, expr, out, f'Argument {i + 1} of {op}') 
        
            # record types represented by _
            ptypes = self.rddl.param_types[name]
            enum_types.update({ptypes[j] for j in underscores})
            
        # types represented by _ must be the same
        if len(enum_types) != 1:
            raise RDDLTypeError(
                f'Sampling parameter(s) _ across all argument(s) of {op} must '
                f'correspond to a single type, got multiple types {enum_types}.\n' + 
                PST(expr, out._current_root))
        
        # objects of type _ must be compatible with sample dimension(s)
        enum_type, = enum_types
        sample_pvar_indices = tuple(scope_pvars[pvar] for pvar in sample_pvars)
        for index in sample_pvar_indices:
            pvar, ptype = objects[index]
            if ptype != enum_type:
                raise RDDLTypeError(
                    f'{op} sampling is performed over type <{enum_type}>, '
                    f'but destination variable <{pvar}> is of type <{ptype}>.\n' + 
                    PST(expr, out._current_root))
             
        out._append(expr, objects, None, True, sample_pvar_indices)

    # ===========================================================================
    # matrix algebra
    # ===========================================================================
    
    def _trace_matrix(self, expr, objects, out):
        _, op = expr.etype
        if op == 'det':
            self._trace_matrix_det(expr, objects, out)
        elif op == 'inverse':
            self._trace_matrix_inv(expr, objects, out, pseudo=False)
        elif op == 'pinverse':
            self._trace_matrix_inv(expr, objects, out, pseudo=True)
        elif op == 'cholesky':
            self._trace_matrix_inv(expr, objects, out, pseudo=False)
        else:
            raise RDDLNotImplementedError(
                f'Internal error: matrix operation {op} is not supported.\n' + 
                PST(expr, out._current_root))
    
    def _trace_matrix_det(self, expr, objects, out):
        * pvars, arg = expr.args
        
        # validate types according to aggregation rule
        iter_objects = [ptype for (_, ptype) in pvars]
        self._check_iteration_variables(objects, iter_objects, expr, out)
        
        # check that matrix is square
        pvar1, pvar2 = iter_objects
        _, ptype1 = pvar1
        _, ptype2 = pvar2
        n1, n2 = self.rddl.object_counts((ptype1, ptype2))      
        if n1 != n2:
            raise RDDLInvalidObjectError(
                f'Matrix in det operation must be square, '
                f'got {n1} objects of type <{ptype1}> '
                f'and {n2} objects of type <{ptype2}>.\n' + 
                PST(expr, out._current_root))
        
        # axes of new free variables in aggregation are appended to value array
        new_objects = objects + iter_objects
        reduced_axes = tuple(range(len(objects), len(new_objects)))   
        cached_sim_info = (new_objects, reduced_axes)
        
        # trace the matrix with the new objects
        self._trace(arg, new_objects, out)
        is_fluent = out.cached_is_fluent(arg)
        RDDLObjectsTracer._check_not_object(
            arg, expr, out, f'Argument of matrix det')     
        
        out._append(expr, objects, None, is_fluent, cached_sim_info)
        
        # log information about matrix operation
        if self.logger is not None:
            message = (f'[info] computing object info for matrix operation:'
                       f'\n\tmatrix operation              =det'
                       f'\n\tdimension variables(s)        ={pvars}'
                       f'\n\tfree object(s) in outer scope ={objects}'
                       f'\n\tfree object(s) in inner scope ={new_objects}'
                       f'\n\treduction axes                ={reduced_axes}\n')
            self.logger.log(message)        
        
    def _trace_matrix_inv(self, expr, objects, out, pseudo):
        _, op = expr.etype
        pvars, arg = expr.args
        
        # check that all pvars are defined in the outer scope        
        scope_pvars = {pvar: i for (i, (pvar, _)) in enumerate(objects)}
        bad_pvars = {pvar for pvar in pvars if pvar not in scope_pvars}
        if bad_pvars:
            raise RDDLInvalidObjectError(
                f'Row or column parameter(s) {bad_pvars} of {op} are not defined '
                f'in outer scope, must be one of {set(scope_pvars.keys())}.\n' + 
                PST(expr, out._current_root))
        
        # check duplicates in pvars
        pvar1, pvar2 = pvars
        if pvar1 == pvar2:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Row and column parameters of {op} must differ, '
                f'got <{pvar1}> and <{pvar2}>.\n' + PST(expr, out._current_root))
            
        # for regular inverse, check that matrix is square
        index_pvar1 = scope_pvars[pvar1]
        index_pvar2 = scope_pvars[pvar2]
        _, ptype1 = objects[index_pvar1]
        _, ptype2 = objects[index_pvar2]
        if not pseudo:
            n1, n2 = self.rddl.object_counts((ptype1, ptype2))        
            if n1 != n2:
                raise RDDLInvalidObjectError(
                    f'Matrix in {op} operation must be square, '
                    f'got {n1} objects of type <{ptype1}> '
                    f'and {n2} objects of type <{ptype2}>.\n' + 
                    PST(expr, out._current_root))
        
        # move the matrix parameters to end of objects
        batch_objects = [pvar for pvar in objects if pvar[0] not in scope_pvars]
        new_objects = batch_objects + [(pvar1, ptype1), (pvar2, ptype2)]
        
        # trace the matrix expression
        self._trace(arg, new_objects, out)
        is_fluent = out.cached_is_fluent(arg)
        RDDLObjectsTracer._check_not_object(
            arg, expr, out, f'Argument of matrix {op}')     
        
        # save the location of the moved indices in objects
        pvar_indices = (index_pvar1, index_pvar2)
        out._append(expr, objects, None, is_fluent, pvar_indices)
        
        # log information about matrix operation
        if self.logger is not None:
            message = (f'[info] computing object info for matrix operation:'
                       f'\n\tmatrix operation              ={op}'
                       f'\n\tdimension variables(s)        ={pvars}'
                       f'\n\tfree object(s) in outer scope ={objects}'
                       f'\n\tfree object(s) in inner scope ={new_objects}'
                       f'\n\tindices in outer scope        ={pvar_indices}\n')
            self.logger.log(message)        



import matplotlib.pyplot as plt
import numpy as np
from PIL import Image

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel
from pyRDDLGym.Visualizer.StateViz import StateViz
from pyRDDLGym import Visualizer


class UAVsVisualizer(StateViz):

    def __init__(self, model: PlanningModel,
                 figure_size=[200, 200],
                 dpi=5,
                 fontsize=8,
                 display=False) -> None:
        self._model = model
        self._states = model.groundstates()
        self._nonfluents = model.groundnonfluents()
        self._objects = model.objects
        self._figure_size = figure_size
        self._dpi = dpi
        self._fontsize = fontsize
        self._interval = 10
        self._asset_path = "/".join(Visualizer.__file__.split("/")[:-1])      
        self._nonfluent_layout = None
        self._state_layout = None
        self._fig, self._ax = None, None
        self._data = None
        self._img = None
    
    def build_nonfluents_layout(self): 
        goal_location = {o: [None, None, None] for o in self._objects['aircraft']}
        
        for k, v in self._nonfluents.items():
            var, objects = self._model.parse(k)
            if var == 'GOAL-X':
                goal_location[objects[0]][0] = v
            elif var == 'GOAL-Y':
                goal_location[objects[0]][1] = v
            elif var == 'GOAL-Z':
                goal_location[objects[0]][2] = v
        
        return {'goal_location': goal_location}
    
    def build_states_layout(self, state):
        drone_location = {o: [None, None, None] for o in self._objects['aircraft']}
        velocity = {o: None for o in self._objects['aircraft']}        
        
        for k, v in state.items():
            var, objects = self._model.parse(k)
            if var == 'pos-x':
                drone_location[objects[0]][0] = v
            elif var == 'pos-y':
                drone_location[objects[0]][1] = v
            elif var == 'pos-z':
                drone_location[objects[0]][2] = v
            elif var == 'vel':
                velocity[objects[0]] = v

        return {'drone_location': drone_location, 'velocity': velocity}

    def init_canvas(self, figure_size, dpi): 
        # plt.style.use('dark_background')
        fig = plt.figure(figsize=figure_size, dpi=dpi)
        ax = fig.add_subplot(projection='3d')

        ax.axes.set_xlim3d(left=-figure_size[0] // 2, right=figure_size[0] // 2) 
        ax.axes.set_ylim3d(bottom=-figure_size[0] // 2, top=figure_size[0] // 2) 
        ax.axes.set_zlim3d(bottom=-figure_size[0] // 2, top=figure_size[0] // 2)

        fig.subplots_adjust(left=0, right=0.1, bottom=0, top=0.1)

        ax.tick_params(labelsize=100)
        
        # plt.axis('scaled')
        # plt.axis('off')

        return fig, ax

    def convert2img(self, fig, ax): 
        ax.set_position((0, 0, 1, 1))
        fig.canvas.draw()

        data = np.frombuffer(fig.canvas.tostring_rgb(), dtype=np.uint8)
        data = data.reshape(fig.canvas.get_width_height()[::-1] + (3,))

        img = Image.fromarray(data)

        self._data = data
        self._img = img

        return img

    def render(self, state):
        self.states = state
        self._fig, self._ax = self.init_canvas(self._figure_size, self._dpi)
    
        nonfluent_layout = self.build_nonfluents_layout()
        state_layout = self.build_states_layout(state)

        self._nonfluent_layout = nonfluent_layout
        self._state_layout = state_layout

        for k, v in nonfluent_layout['goal_location'].items():
            self._ax.plot([v[0]], [v[1]], [v[2]],
                          color='seagreen', marker='X', markersize=200)
        
        for k, v in state_layout['drone_location'].items():
            self._ax.plot([v[0]], [v[1]], [v[2]],
                          color='deepskyblue', marker='>', markersize=200)

        img = self.convert2img(self._fig, self._ax)

        self._ax.cla()
        plt.cla()
        plt.close()

        return img
    
    def gen_inter_state(self, beg_state, end_state, steps):
        state_buffer = []
        beg_x = beg_state['xPos']
        beg_y = beg_state['yPos']
        end_x = end_state['xPos']
        end_y = end_state['yPos']

        if (beg_x, beg_y) == (end_x, end_y):
            state_buffer = [beg_state, end_state]
        else:
            for i in range(steps):
                x = beg_x + 1 / steps * (end_x - beg_x) * i
                y = beg_y + 1 / steps * (end_y - beg_y) * i
                state = beg_state.copy()
                state['xPos'] = x
                state['yPos'] = y
                state_buffer.append(state)
            state_buffer.append(end_state)

        return state_buffer
    



from typing import Set, Sized, Tuple

import sympy as sp
from xaddpy.xadd import XADD
from xaddpy.xadd.xadd import ControlFlow, DeltaFunctionSubstitution

from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel, RDDLGroundedModel
from pyRDDLGym.Core.ErrorHandling.RDDLException import (
    RDDLInvalidNumberOfArgumentsError,
    RDDLNotImplementedError,
    RDDLTypeError,
)
from pyRDDLGym.Core.Parser.expr import Expression
from pyRDDLGym.XADD.helper import get_bernoulli_node_id


VALID_RELATIONAL_OPS = {'>=', '>', '<=', '<', '==', '~='}
OP_TO_XADD_OP = {
    '*': 'prod', '+': 'add', '-': 'subtract', '/': 'div',
    '|': 'or', '^': 'and', '~=': '!='
}
PRIME = '\''
UNIFORM_VAR_NAME = '#_UNIFORM_{num}'
GAUSSIAN_VAR_NAME = '#_GAUSSIAN_{num}'
EXPONENTIAL_VAR_NAME = '#_EXPONENTIAL_{num}'


class RDDLModelWXADD(PlanningModel):

    def __init__(
            self,
            model: RDDLGroundedModel,
            context: XADD = None,
            simulation: bool = True,
    ):
        super().__init__()
        self.model = model
        self._context: XADD = XADD() if context is None else context
        self._var_name_to_node_id = {}
        self._sympy_var_to_node_id = {}
        self._sympy_var_name_to_var_name = {}
        self._var_name_to_sympy_var_name = {}
        self._op_to_node_id = {}
        self._node_id_to_op = {}
        self._curr_pvar = None
        self.rvs = self._context._random_var_set

        # Get grounded variables
        self.states = model.states
        self.actions = model.actions
        self.nonfluents = model.nonfluents
        self.next_state = model.next_state
        self.prev_state = model.prev_state
        self.init_state = model.init_state
        self.cpfs = model.cpfs
        self.cpforder = model.cpforder
        self.gvar_to_cpforder = model.gvar_to_cpforder
        self.reward = model.reward
        self.terminals = model.terminals
        self.preconditions = model.preconditions
        self.invariants = model.invariants
        self.derived = model.derived
        self.interm = model.interm
        self.observ = model.observ
        self.observranges = model.observranges
        self.objects = model.objects
        self.actionsranges = model.actionsranges
        self.statesranges = model.statesranges
        self.gvar_to_type = model.gvar_to_type
        self.pvar_to_type = model.pvar_to_type
        self.gvar_to_pvar = model.gvar_to_pvar
        self.max_allowed_actions = model.max_allowed_actions
        self.horizon = model.horizon
        self.discount = model.discount
        self.compiled = False
        self.simulation = simulation
        self._need_postprocessing = False
        self._postprocessing_kwargs = {}

    def compile(self, simulation=None):
        if simulation is not None and isinstance(simulation, bool):
            self.simulation = simulation
        self.reset_dist_var_num()
        self.convert_cpfs_to_xadds()
        self.compiled = True

    def reset_dist_var_num(self):
        self._num_uniform = 0
        self._num_gaussian = 0
        self._num_exponential = 0

    def convert_cpfs_to_xadds(self):
        
        # the API for accessing CPFs is now dict of {name: (objects, expr)}
        # see Mike pull request from Dec 18
        cpfs = {name: expr for name, (_, expr) in self.cpfs.items()}
        cpfs = sorted(cpfs)
        
        # Handle state-fluent
        for cpf in cpfs:
            self._need_postprocessing = False
            _, expr = self.cpfs[cpf]
            pvar_name = self.gvar_to_pvar[cpf]
            if pvar_name != self._curr_pvar:
                self._curr_pvar = pvar_name
                self.reset_dist_var_num()
            expr_xadd_node_id = self.expr_to_xadd(expr)
            # Post-processing for Bernoulli CPFs
            if self._need_postprocessing:
                expr_xadd_node_id = self.postprocess(
                    expr_xadd_node_id,
                    **self._postprocessing_kwargs)
            self.cpfs[cpf] = expr_xadd_node_id
        self.cpfs = self.cpfs

        # Reward
        expr = self.reward
        expr_xadd_node_id = self.expr_to_xadd(expr)
        self.reward = expr_xadd_node_id

        # Terminal condition
        terminals = []
        for i, terminal in enumerate(self.terminals):
            expr = terminal
            expr_xadd_node_id = self.expr_to_xadd(expr)
            terminals.append(expr_xadd_node_id)
        self.terminals = terminals

        # Skip preconditions
        # preconditions = []
        # for i, precondition in enumerate(self.preconditions):
        #     expr = precondition
        #     expr_xadd_node_id = self.expr_to_xadd(expr)
        #     preconditions.append(expr_xadd_node_id)
        # self.preconditions = preconditions
        
        # Also skip invariants
        # invariants = []
        # for i, invariant in enumerate(self.invariants):
        #     expr = invariant
        #     expr_xadd_node_id = self.expr_to_xadd(expr)
        #     invariants.append(expr_xadd_node_id)
        # self.invariants = invariants      

    def expr_to_xadd(self, expr: Expression) -> int:
        node_id = self._op_to_node_id.get(expr)
        if node_id is not None:
            return node_id

        etype, op = expr.etype
        if etype == "constant":
            node_id = self.constant_to_xadd(expr)
        elif etype == "pvar":
            node_id = self.pvar_to_xadd(expr)
        elif etype == "aggregation":
            node_id = self.aggr_to_xadd(expr)
        elif etype == "control":
            node_id = self.control_to_xadd(expr)
        elif etype == "randomvar":
            node_id = self.randomvar_to_xadd(expr)
        elif etype == "func":
            node_id = self.func_to_xadd(expr)
        elif etype == "arithmetic":
            node_id = self.arithmetic_to_xadd(expr)
        elif etype == "relational":
            node_id = self.relational_to_xadd(expr)
        elif etype == "boolean":
            node_id = self.bool_to_xadd(expr)
        else:
            raise Exception(f'Internal error: type {etype} is not supported.')
        node_id = self._context.make_canonical(node_id)
        return node_id

    def constant_to_xadd(self, expr: Expression) -> int:
        assert expr.etype[0] == 'constant'
        const = sp.sympify(expr.args, locals=self.ns)
        return self._context.convert_to_xadd(const)

    def pvar_to_xadd(self, expr: Expression) -> int:
        assert expr.etype[0] == 'pvar'
        var, args = expr.args
        var_type = self.gvar_to_type[var]
        if var in self.nonfluents:
            var_ = self.nonfluents[var]
            node_id = self._context.convert_to_xadd(sp.S(var_))
            self._var_name_to_node_id[var] = node_id            
        else:
            var_, node_id = self.add_sympy_var(var, var_type)
        return node_id

    def add_sympy_var(
            self,
            var_name: str,
            var_type: str,
            random: bool = False,
            **kwargs,
    ) -> Tuple[sp.Symbol, int]:
        if var_name in self._var_name_to_sympy_var_name:
            var_ = self.ns[var_name]
            node_id = self._var_name_to_node_id[var_name]
            return var_, node_id
        var_ = self.ns.setdefault(
            var_name,
            sp.Symbol(
                var_name.replace('-', '_'),
                bool=var_type == 'bool',
                random=random
            )
        )
        node_id = self._context.convert_to_xadd(var_, **kwargs)
        self._sympy_var_to_node_id[var_] = node_id
        self._var_name_to_node_id[var_name] = node_id
        self._sympy_var_name_to_var_name[str(var_)] = var_name
        self._var_name_to_sympy_var_name[var_name] = str(var_)
        return var_, node_id

    def aggr_to_xadd(self, expr: Expression) -> int:
        """
        For lifted variables, we cannot evaluate aggregations such as 
            'exists', 'forall', 'sum', 'prod', 'avg', 'minimum', 'maximum'.
        Thus, these will be evaluated explicitly when grounding. 
        For now, it is sufficient to return a node with the right type.
        Additionally, we need to know what variables are given as arguments to
        this expression.
        
        For example, for an 'exists' operation,
            1) create a unique sympy variable associated with this expression
                The variable name should be unique but should be identical for 
                identical operations). 
                An exists operation can be uniquely identified by 
                    * the argument(s) over which we aggregate
                        these are ?x and ?y from exists_{?x: xpos, ?y: ypos}
                    * the operand... Can we recognize identical expressions?
                
            2) 
        """
        assert expr.etype[0] == 'aggregation'
        etype, op = expr.etype
        args = expr.args
        arg_vars = []
        for arg in args:
            if arg[0] == 'typed_var':
                arg_vars += list(arg[1])  # TODO: handle these
        assert len(arg_vars) >= 2

        num_aggregations = len(self._aggr_to_scope)
        postfix = '__' + '__'.join([a for a in arg_vars[::2]])

        # Aggregations that return Booleans
        if op == 'exists' or op == 'forall':
            var_sp = sp.Symbol(f'exists{postfix}_{num_aggregations}', bool=True)

        # Aggregations that return int/float values
        elif op in ('sum', 'prod', 'avg', 'minimum', 'maximum'):
            var_sp = sp.Symbol(f'{op}{postfix}_{num_aggregations}')
        else:
            raise RDDLNotImplementedError

        # TODO: Scope doesn't contain alls
        self._aggr_to_scope[var_sp] = [self.ns[var.split('/')[0]] for var in expr.scope]
        node_id = self._context.get_leaf_node(var_sp)
        return node_id        

    def control_to_xadd(self, expr: Expression) -> int:
        """
        Control corresponds to if-else if-else statements.
        For ifs and else ifs, the first argument is the condition, 
        the second is the true branch, and the third argument is the false branch
        (else can be absorbed into the false branch of the last else if statement).
        
        Let's say the condition is represented by an XADD node n1;
        the true branch n2; and the false branch n3.
        Then, the final node can be obtained by the leaf operation ControlFlow,
        which goes to the leaf nodes of n1 and creates a decision node whose 
        high branch corresponds to the node n2 and whose low branch is the node n3.
        """
        assert expr.etype[0] == 'control'
        args = list(map(self.expr_to_xadd, expr.args))
        condition = args[0]
        true_branch = args[1]
        false_branch = args[2]
        leaf_op = ControlFlow(
            true_branch=true_branch,
            false_branch=false_branch,
            context=self._context
        )
        node_id = self._context.reduce_process_xadd_leaf(
            condition,
            leaf_op=leaf_op,
            decisions=[],
            decision_values=[]
        )
        return node_id

    def randomvar_to_xadd(self, expr: Expression) -> int:
        """Converts a random variable expression to an XADD node.

        When `self.simulation` is True, we reparameterize RVs for sampling purposes. 
          That is, a Bernoulli rv can be sampled by sampling a uniform random value 
          within the range [0, 1] and check whether the sampled value is greater/smaller 
          than the Bernoulli parameter.
        On the other hand, when `self.simulation` is False, we will create a node 
            something like the following:
                ([Bernoulli_params_1_123])
            where `123` is the node ID associated with the Bernoulli parameter,
            and `1` denotes the number of parameters, which is introduced to help with
            parsing later in a downstream task.

        Currently, `self.simulation` should be set to True for all rvs except for 
        Bernoulli. Furthermore, we assume that the Bernoulli node is used only at a
        leaf node of the XADD.
        """
        assert expr.etype[0] == 'randomvar'
        dist = expr.etype[1]
        args = list(map(self.expr_to_xadd, expr.args))
        
        if dist != 'Bernoulli':
            assert self.simulation, ("Currently, only Bernoulli" 
                                     "can be used for non-simulation.")

        if dist == 'Bernoulli':
            assert len(args) == 1
            proba = args[0]
            # Sample a Bernoulli rv by sampling a uniform rv
            if self.simulation:
                num_rv = self._num_uniform
                unif_rv = sp.Symbol(UNIFORM_VAR_NAME.format(num=num_rv), random=True)
                uniform = self._context.convert_to_xadd(
                    unif_rv,
                    params=(0, 1),  # rv ~ Uniform(0, 1)
                    type='UNIFORM'
                )
                node_id = self._context.apply(uniform, proba, '<=')
                self._num_uniform += 1
            # Create a Bernoulli node
            # TODO: How to assert that a Bernoulli node can only come at a leaf?
            else:
                name = f'{dist}_params_{len(args)}_{proba}'
                rv, node_id = self.add_sympy_var(name,
                                                 'real',
                                                 random=True,
                                                 params=(proba,),
                                                 type='BERNOULLI')
                # Need to postprocess the node to make leaf nodes 
                # represent Bernoulli probabilities
                self._need_postprocessing = True
            return node_id
        
        elif dist == 'Binomial':
            pass
        
        elif dist == 'krondelta':
            # Change Boolean leaf values to 0s and 1s
            assert len(args) == 1
            arg, = args
            node_id = self._context.unary_op(arg, op='int')
            return node_id

        elif dist == 'exponential':
            assert len(args) == 1
            num_rv = self._num_uniform
            unif_rv = sp.Symbol(UNIFORM_VAR_NAME.format(num=num_rv), random=True)
            uniform = self._context.convert_to_xadd(
                unif_rv,
                params=(0, 1),  # rv ~ Uniform(0, 1)
                type='UNIFORM'
            )
            # '-log(1 - U) * scale' is an exponential sample reparameterized with Uniform
            minus_u = self._context.unary_op(uniform, '-')
            log1_minus_u = self._context.unary_op(minus_u, 'log1p')
            neg_log1_minus_u = self._context.unary_op(log1_minus_u, '-')
            scale = args[0]
            node_id = self._context.apply(neg_log1_minus_u, scale, 'prod')
            self._num_uniform += 1
            return node_id
        
        elif dist == 'normal':
            assert len(args) == 2
            mean, var = args
            num_rv = self._num_gaussian
            gauss_rv = sp.Symbol(GAUSSIAN_VAR_NAME.format(num=num_rv), random=True)
            gaussian = self._context.convert_to_xadd(
                gauss_rv,
                params=(0, 1),  # rv ~ Normal(0, 1)
                type='NORMAL',
            )
            # mean + sqrt(var) * epsilon
            std = self._context.unary_op(var, 'sqrt')
            scaled = self._context.apply(std, gaussian, 'prod')
            node_id = self._context.apply(mean, scaled, 'add')
            self._num_gaussian += 1
            return node_id
        
        else:
            raise RDDLNotImplementedError(
                f'Distribution {dist} does not allow reparameterization'
            )  # TODO: print stack trace?
            
        return
    
    def func_to_xadd(self, expr: Expression) -> int:
        assert expr.etype[0] == 'func'
        etype, op = expr.etype
        args = list(map(self.expr_to_xadd, expr.args))
        
        if op == 'pow':
            assert len(args) == 2
            pow = expr.args[1].value
            node_id = self._context.unary_op(args[0], op, pow)
            
        elif op == 'max' or op == 'min':
            assert len(args) == 2
            node_id = self._context.apply(args[0], args[1], op)
            
        else:
            assert len(args) == 1
            node_id = self._context.unary_op(args[0], op)
            
        node_id = self._context.make_canonical(node_id)
        return node_id

    def arithmetic_to_xadd(self, expr: Expression) -> int:
        assert expr.etype[0] == 'arithmetic'
        etype, op = expr.etype
        args = list(map(self.expr_to_xadd, expr.args))
        if len(args) == 1:  # Unary operation
            node_id = self._context.unary_op(args[0], op)
        elif len(args) == 2:
            node_id = self._context.apply(args[0], args[1], OP_TO_XADD_OP.get(op, op))
        elif len(args) > 2:
            node_id = args[0]
            for arg in args[1:]:
                node_id = self._context.apply(node_id, arg, OP_TO_XADD_OP.get(op, op))
        else:
            raise ValueError("Operations with XADD nodes should be unary or binary")
        return node_id

    def relational_to_xadd(self, expr: Expression) -> int:
        assert expr.etype[0] == 'relational'
        etype, op = expr.etype
        if op not in VALID_RELATIONAL_OPS:
            raise RDDLNotImplementedError(
                f'Relational operator {op} is not supported: must be one of {VALID_RELATIONAL_OPS}'
            )  # TODO: print stack trace?

        args = list(map(self.expr_to_xadd, expr.args))            
        if not isinstance(args, Sized):
            raise RDDLTypeError(
                f'Internal error: expected Sized, got {type(args)}'
            )  # TODO: print stack trace?
        elif len(args) != 2:
            raise RDDLInvalidNumberOfArgumentsError(
                f'Relational operator {op} requires 2 args, got {len(args)}'
            )  # TODO: print stack trace?
        
        node_id = self._context.apply(args[0], args[1], OP_TO_XADD_OP.get(op, op))
        return node_id

    def bool_to_xadd(self, expr: Expression) -> int: 
        assert expr.etype[0] == 'boolean'
        etype, op = expr.etype
        args = list(map(self.expr_to_xadd, expr.args))
        if len(args) == 1 and op == '~':
            node_id = self._context.unary_op(args[0], OP_TO_XADD_OP.get(op, op))
            return node_id
        elif len(args) >= 2:
            if op == '|' or op == '^':
                node_id = args[0]
                for arg in args[1:]:
                    node_id = self._context.apply(node_id, arg, OP_TO_XADD_OP.get(op, op))
                return node_id
            elif len(args) == 2:
                if op == '~':
                    return  # When does this happen? 
                elif op == '=>':
                    return
                elif op == '<=>':
                    return
        
        raise RDDLInvalidNumberOfArgumentsError(
            f'Logical operator {op} does not have the required number of args, got {len(args)}' + 
            f'\n{expr}'  # TODO: print stack trace?
        )

    def postprocess(self, node_id: int, **kwargs) -> int:
        """Post processes a CPF XADD node.
        
        Currently, this method exists solely to handle the Bernoulli
        CPFs when the simulation flag is set to False. In this case,
        we replace the Bernoulli leaf nodes with their corresponding
        parameter nodes.
        """
        var_set = self._context.collect_vars(node_id)
        bernoulli_set = set()
        for v in var_set:
            name = str(v)
            if name.startswith('Bernoulli'):
                bernoulli_set.add(v)
        for rv in bernoulli_set:
            bern_param_node_id = get_bernoulli_node_id(rv)
            leaf_op = DeltaFunctionSubstitution(
                sub_var=rv, xadd_sub_at_leaves=node_id, context=self._context)
            node_id = self._context.reduce_process_xadd_leaf(node_id=bern_param_node_id,
                                                             leaf_op=leaf_op,
                                                             decisions=[],
                                                             decision_values=[])
        try:
            node_id = self._context.reduce_lp(node_id)
        except:
            pass
        return node_id

    @property
    def ns(self):
        if not hasattr(self, '_ns'):
            self._ns = {}
        return self._ns

    def print(self, node_id):
        print(self._context.get_exist_node(node_id))

    def collect_vars(self, node_id: int) -> Set[str]:
        """Returns the set containing variables existing in the current node (in str type)"""
        var_set = self._context.collect_vars(node_id)
        var_set = var_set.difference(self._context._random_var_set)
        return set(self._sympy_var_name_to_var_name[str(v)] for v in var_set)
    
    @property
    def vars_in_rew(self) -> Set[str]:
        if not hasattr(self, '_vars_in_rew'):
            self._vars_in_rew = self.collect_vars(self.reward)
        return self._vars_in_rew
    
    @vars_in_rew.setter
    def vars_in_rew(self, var: Set[str]):
        self._vars_in_rew = var

    @property
    def simulation(self) -> bool:
        return self._simulation

    @simulation.setter
    def simulation(self, sim: bool):
        self._simulation = sim

    @property
    def compiled(self) -> bool:
        return self._compiled

    @compiled.setter
    def compiled(self, comp: bool):
        self._compiled = comp



import jax
import jax.numpy as jnp
import jax.random as random
import matplotlib
import matplotlib.pyplot as plt
import numpy as np
from typing import Callable, Dict, Tuple

from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Jax.JaxRDDLCompiler import JaxRDDLCompiler
from pyRDDLGym.Core.Jax.JaxRDDLBackpropPlanner import JaxRDDLCompilerWithGrad
from pyRDDLGym.Core.Jax.JaxRDDLLogic import FuzzyLogic


class Subplots:
    
    def __init__(self, num_plots: int,
                 figsize: Tuple[int, int]) -> None:
        self.num_plots = num_plots
        
        # sub-plots arranged in a square grid
        self.dim = dim = np.floor(np.sqrt(num_plots)).astype(int) + 1
        w, h = figsize
        self.figure, self.axes = plt.subplots(
            nrows=dim, ncols=dim, figsize=(w * dim, h * dim))
        self.ix, self.iy = 0, 0
    
    def current_axis(self) -> matplotlib.axis.Axis:
        return self.axes[self.iy, self.ix]
    
    def next_axis(self) -> None:
        self.ix += 1
        if self.ix >= self.dim:
            self.ix = 0
            self.iy += 1
        
    def blank_unused_axes(self) -> None:
        while self.iy < self.dim:
            self.current_axis().set_axis_off()
            self.next_axis()
            
    def close(self) -> None:
        plt.clf()
        plt.close(self.figure)
        del self.axes
        del self.figure
    
    
class JaxRDDLModelError:
    '''Assists in the analysis of model error that arises from replacing exact
    operations with fuzzy logic.
    '''

    def __init__(self, rddl: RDDLLiftedModel,
                 test_policy: Callable,
                 batch_size: int,
                 rollout_horizon: int=None,
                 logic: FuzzyLogic=FuzzyLogic()) -> None:
        '''Creates a new instance to empirically estimate the model error.
        
        :param rddl: the RDDL domain to optimize
        :param test_policy: policy to generate rollouts
        :param batch_size: batch size
        :param rollout_horizon: lookahead planning horizon: None uses the
        horizon parameter in the RDDL instance
        :param logic: a subclass of FuzzyLogic for mapping exact mathematical
        operations to their differentiable counterparts
        '''
        self.rddl = rddl
        self.test_policy = test_policy
        self.batch_size = batch_size
        if rollout_horizon is None:
            rollout_horizon = rddl.horizon
        self.horizon = rollout_horizon
        self.logic = logic
        
        self._jax_compile()
    
    def _jax_compile(self):
        rddl = self.rddl
        n_batch = self.batch_size
        test_policy = self.test_policy
        
        # Jax compilation of the differentiable RDDL for training
        self.train_compiled = JaxRDDLCompilerWithGrad(
            rddl=rddl,
            logic=self.logic)
        self.train_compiled.compile()
        
        # Jax compilation of the exact RDDL for testing
        self.test_compiled = JaxRDDLCompiler(rddl=rddl)
        self.test_compiled.compile()
        
        # roll-outs
        def _jax_wrapped_train_policy(key, policy_params, step, subs):
            test_actions = test_policy(key, policy_params, step, subs)
            train_actions = jax.tree_map(
                lambda x: x.astype(self.test_compiled.REAL),
                test_actions)
            return train_actions
        
        train_rollouts = self.train_compiled.compile_rollouts(
            policy=_jax_wrapped_train_policy,
            n_steps=self.horizon,
            n_batch=n_batch)
        
        test_rollouts = self.test_compiled.compile_rollouts(
            policy=test_policy,
            n_steps=self.horizon,
            n_batch=n_batch)
        
        # batched initial subs
        def _jax_wrapped_batched_subs(subs): 
            train_subs, test_subs = {}, {}
            for (name, value) in subs.items():
                value = jnp.asarray(value)[jnp.newaxis, ...]
                train_value = jnp.repeat(value, repeats=n_batch, axis=0)
                train_value = train_value.astype(self.test_compiled.REAL)
                train_subs[name] = train_value
                test_subs[name] = jnp.repeat(value, repeats=n_batch, axis=0)
            for (state, next_state) in rddl.next_state.items():
                train_subs[next_state] = train_subs[state]
                test_subs[next_state] = test_subs[state]
            return train_subs, test_subs
        
        # forward simulation
        def _jax_wrapped_simulate_particles(
                key, policy_params, subs, params_train, params_test):
            
            # generate rollouts from train and test model for the same inputs
            train_subs, test_subs = _jax_wrapped_batched_subs(subs)
            train_log = train_rollouts(key, policy_params, train_subs, params_train)
            test_log = test_rollouts(key, policy_params, test_subs, params_test)
            
            # extract CPF and reward trajectories
            train_fluents, test_fluents = {}, {}
            train_fluents['reward'] = train_log['reward']
            test_fluents['reward'] = test_log['reward']
            for cpf in self.train_compiled.cpfs:
                train_fluents[cpf] = train_log['pvar'][cpf]
                test_fluents[cpf] = test_log['pvar'][cpf]
            return train_fluents, test_fluents
        
        self.simulation = jax.jit(_jax_wrapped_simulate_particles)
            
    def sensitivity(self, key: random.PRNGKey,
                    policy_params: Dict,
                    param_ranges=[1., 5., 10., 50., 100., 500., 1000.],
                    bars=100,
                    epochs=5,
                    subs: Dict=None,
                    figsize: Tuple[int, int]=(4, 2)) -> None:
        if subs is None:
            subs = self.test_compiled.init_values
        test_params = self.test_compiled.model_params
        
        # subplots
        param_ranges = list(param_ranges)
        num_rows = len(param_ranges)
        num_cols = epochs
        w, h = figsize
        
        # collect data
        train_dicts, test_dicts = [], []
        for weight in param_ranges:
            model_params = {name: weight 
                            for name in self.train_compiled.model_params}
            key, subkey = random.split(key)
            train_fluents, test_fluents = self.simulation(
                subkey, policy_params, subs, model_params, test_params)  
            train_fluents = {k: np.asarray(v, dtype=np.float64).reshape(
                                (v.shape[0], v.shape[1], -1))
                             for (k, v) in train_fluents.items()}
            test_fluents = {k: np.asarray(v, dtype=np.float64).reshape(
                                (v.shape[0], v.shape[1], -1))
                            for (k, v) in test_fluents.items()}
            train_dicts.append(train_fluents)
            test_dicts.append(test_fluents)
        
        # select epochs as equally spaced as possible including 0 and T - 1
        num_epochs = train_fluents['reward'].shape[1]
        iepochs = np.round(np.linspace(0, num_epochs - 1, epochs)).astype(int)
        
        # plot for cumulative reward
        fig, axes = plt.subplots(
            nrows=num_rows, ncols=1, figsize=(w * 1.5, h * num_rows),
            sharex='col')      
        for (row, weight) in enumerate(param_ranges):
            train_data = train_dicts[row]['reward'][:,:, ...]
            test_data = test_dicts[row]['reward'][:,:, ...]
            train = np.ravel(np.sum(train_data, axis=1))
            test = np.ravel(np.sum(test_data, axis=1))
            xmin = min(np.min(train), np.min(test))
            xmax = max(np.max(train), np.max(test))
            bins = np.linspace(xmin, xmax, bars)
            axis = axes[row]
            axis.hist(train, bins=bins, alpha=0.5, label='train', histtype='step')
            axis.hist(test, bins=bins, alpha=0.5, label='test', histtype='step')
            axis.legend()
            axis.set_xlabel(f'cuml reward')
            axis.set_ylabel(f'weight {weight}')
        plt.tight_layout()
        fig.savefig(f'sensitivity_cuml_reward.pdf')
        plt.clf()
        plt.close(fig)
        
        # each plot per grounded fluent, rows weights, columns epochs
        for name in train_fluents:
            grounded_names = list(self.rddl.ground_names(
                name, self.rddl.param_types.get(name, [])))
            for (iname, gname) in enumerate(grounded_names):
                fig, axes = plt.subplots(
                    nrows=num_rows, ncols=num_cols,
                    figsize=(w * num_cols, h * num_rows), sharex='col')            
                for (row, weight) in enumerate(param_ranges):
                    train_data = train_dicts[row][name][:,:, iname]
                    test_data = test_dicts[row][name][:,:, iname]
                    for (col, epoch) in enumerate(iepochs):
                        train = train_data[:, epoch]
                        test = test_data[:, epoch]
                        xmin = min(np.min(train), np.min(test))
                        xmax = max(np.max(train), np.max(test))
                        bins = np.linspace(xmin, xmax, bars)
                        axis = axes[row, col]
                        axis.hist(train, bins=bins, 
                                  alpha=0.5, label='train', histtype='step')
                        axis.hist(test, bins=bins, 
                                  alpha=0.5, label='test', histtype='step')
                        axis.set_xlabel(f'epoch {epoch}')
                        axis.set_ylabel(f'weight {weight}')
                        if row == 0 and col == 0:
                            axis.legend()
                print(f'ploting fluent {gname}...')
                plt.tight_layout()
                fig.savefig(f'sensitivity_{gname}.pdf')
                plt.clf()
                plt.close(fig)
    
    def summarize(self, key: random.PRNGKey,
                  policy_params: Dict,
                  subs: Dict=None,
                  figsize: Tuple[int, int]=(6, 3)) -> None:
        
        # get train and test particles
        if subs is None:
            subs = self.test_compiled.init_values
        train_fluents, test_fluents = self.simulation(
            key, policy_params, subs,
            self.train_compiled.model_params,
            self.test_compiled.model_params)
    
        # figure out # of plots required to plot each fluent component separately
        sizes = {name: np.prod(value.shape[2:], dtype=int)
                 for (name, value) in train_fluents.items()}
        
        # plot each fluent component
        subplots = Subplots(sum(sizes.values()), figsize)
        for (name, train_value) in train_fluents.items():
            train_value = np.asarray(train_value, dtype=np.float64)
            test_value = np.asarray(test_fluents[name], dtype=np.float64)
            
            # calculate statistics for fluent
            train_value = train_value.reshape(*train_value.shape[:2], -1, order='C')
            test_value = test_value.reshape(*test_value.shape[:2], -1, order='C')
            train_mean = np.mean(train_value, axis=0)
            test_mean = np.mean(test_value, axis=0)
            train_std = np.std(train_value, axis=0)
            test_std = np.std(test_value, axis=0)     
            
            grounded_names = list(self.rddl.ground_names(
                name, self.rddl.param_types.get(name, [])))
            assert len(grounded_names) == sizes[name]
            
            # plot each component j of fluent on a separate plot
            for j in range(sizes[name]):
                axis = subplots.current_axis()
                
                # show mean
                xs = np.arange(train_mean.shape[0])
                axis.plot(xs, train_mean[:, j], label='approx mean', color='blue')
                axis.plot(xs, test_mean[:, j], label='true mean',
                          linestyle='dashed', color='black')
                
                # show two standard deviations spread
                axis.fill_between(xs, train_mean[:, j] - 2 * train_std[:, j],
                                  train_mean[:, j] + 2 * train_std[:, j],
                                  color='blue', alpha=0.15)
                axis.fill_between(xs, test_mean[:, j] - 2 * test_std[:, j],
                                  test_mean[:, j] + 2 * test_std[:, j],
                                  color='black', alpha=0.15)
                
                axis.legend()
                axis.set_xlabel('epoch')
                axis.set_ylabel(grounded_names[j])
                subplots.next_axis()                
        subplots.blank_unused_axes()        
        plt.tight_layout()
        subplots.figure.savefig('summary.pdf')
        subplots.close()



import base64
import json
import socket
import xml.etree.ElementTree as xmltree

from pyRDDLGym import RDDLEnv
from pyRDDLGym.Core.Compiler.RDDLModel import PlanningModel

class RDDLSimAgent:
    ''' creates a TCP/IP server that listens to the provided port and passes
    messages between a pyRDDLGym environment and a client that is
    designed to interact with rddlsim (https://github.com/ssanner/rddlsim)'''

    def __init__(self, domain, instance, numrounds, time, port=2323):

        # concatenate domain and instance files
        print(f"loading domain {domain}...", flush=True)
        f = open(domain)
        self.task = f.read()
        f.close()
        print(f"loading instance {instance}...", flush=True)
        f = open(instance)
        self.task = self.task + f.read()
        f.close()

        # encode task
        print(f"encoding task for sharing in TCP connections...", flush=True)
        self.task = base64.b64encode(str.encode(self.task))
        self.task = self.task.decode("ascii")
        
        # create RDDLEnv
        print(f"creating RDDL environment...", flush=True)
        self.env = RDDLEnv.RDDLEnv(domain=domain, instance=instance)
        
        # initialize RDDLSimAgent
        self.roundsleft = numrounds
        self.currentround = 0
        self.time = time
        self.address = ("127.0.0.1", port)
        self.client = ""
        self.problem = ""
        self.total_reward = 0.0

        # Data in case there is a dump request
        self.logs = []

    def run(self):
        ''' starts the RDDLSimAgent to wait for a planner to connect'''

        print(f"establishing socket...", flush=True)
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        with sock:

            # Force the connection to this port (sometimes it stays locked after repeated runs).
            # https://stackoverflow.com/questions/4465959/python-errno-98-address-already-in-use
            print(f"forcing connection...", flush=True)
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)

            sock.bind(self.address)
            print(f"listening at address {self.address[0]} with port {self.address[1]}...", flush=True)
            sock.listen(1)
            connection, client_address = sock.accept()
            with connection:
                self.run_session(connection)
            connection.close()
        sock.close()
        self.env.close()

    def run_session(self, connection):
        ''' runs an interactive session between the pyRDDLGym environment
        and a connected rddlsim client and terminates afterwards'''

        # handle session request
        data = self.receive_message(connection)
        self.process_init_session_request(data)
        print(f"session request from {self.client} for {self.problem}", flush=True)
        msg = self.build_session_request_msg()
        self.send_message(connection, msg)
        session_request_expected = False
        print("session initialized", flush=True)

        while self.roundsleft > 0:
            self.run_round(connection)

        msg = self.build_session_end_msg()
        self.send_message(connection, msg)

    def run_round(self, connection):

        self.logs.append([])

        # handle round request
        data = self.receive_message(connection)
        self.process_round_request(data)
        print(f"starting round {self.currentround}", flush=True)
        msg = self.build_round_request_msg()
        self.send_message(connection, msg)

        # initialize round
        state = self.env.reset()
        round_reward = 0.0
        turn = 1
        msg = self.build_state_msg(state, turn, 0.0)
        self.send_message(connection, msg)

        # run round
        while True:
            data = self.receive_message(connection)
            actions = self.process_action(data)

            self.logs[-1].append({
                "state": json.loads(str(state).replace('\'', '"').replace('True', 'true').replace('False', 'false')),
                "actions": json.loads(str(actions).replace('\'', '"').replace('"true"', 'true')),
            })

            next_state, reward, done, info = self.env.step(actions)

            self.logs[-1][-1]["reward"] = float(reward)

            round_reward += reward
            state = next_state

            turn = turn + 1
            if turn == self.env.horizon:
                msg = self.build_round_end_msg(reward, round_reward)
                self.send_message(connection, msg)
                self.total_reward += round_reward

                self.logs[-1].append({
                    "reward": float(round_reward),
                    "state": json.loads(str(state).replace('\'', '"').replace('True', 'true').replace('False', 'false')),
                    "actions": False,
                })

                break
            else:
                msg = self.build_state_msg(state, turn, 0.0)
                self.send_message(connection, msg)

    def dump_data(self, fn):
        """Dumps the data to a json file"""
        with open(fn, "w") as f:
            json.dump(self.logs, f)

    def send_message(self, connection, msg):
        #print(f"sending message: {msg}")
        connection.send(str.encode(msg))

    def receive_message(self, connection):
        data = connection.recv(8192)
        if data:
            data = data.decode('UTF-8')
            data = data[:-1]
            #print(f"received message: {data}")
        else:
            print("Error: connection lost", flush=True)
            exit(1)
        return data

    def build_session_request_msg(self):
        msg = "<session-init>"
        msg = msg + "<task>" + str(self.task) + "</task>"
        msg = msg + "<session-id>0</session-id>"
        msg = msg + "<num-rounds>" + str(self.roundsleft) + "</num-rounds>"
        msg = msg + "<time-allowed>" + str(self.time) + "</time-allowed>"
        msg = msg + "</session-init>"
        return msg

    def build_round_request_msg(self):
        msg = "<round-init>"
        msg = msg + "<round-num>" + str(self.currentround) + "</round-num>"
        msg = msg + "<time-left>1000</time-left>"
        msg = msg + "<rounds-left>" + str(self.roundsleft) + "</rounds-left>"
        msg = msg + "<sessionID>0</sessionID>"
        msg = msg + "</round-init>"
        return msg

    def build_state_msg(self, state, turn, rew):
        #print(state)
        msg = "<turn>"
        msg = msg + "<turn-num>" + str(turn) + "</turn-num>"
        msg = msg + "<time-left>1000</time-left>"
        msg = msg + "<immediate-reward>" + str(rew) + "</immediate-reward>"
        for key in state:
            msg = msg + "<observed-fluent>"
            fluent_name = key.split(PlanningModel.FLUENT_SEP)[0]
            objects = key.split(PlanningModel.FLUENT_SEP)[1:]
            objects = PlanningModel.FLUENT_SEP.join(objects)
            objects = objects.split(PlanningModel.OBJECT_SEP)
            var = key.split("_")
            msg = msg + "<fluent-name>" + fluent_name + "</fluent-name>"
            for object in objects:
                msg = msg + "<fluent-arg>" + object + "</fluent-arg>"
            msg = msg + "<fluent-value>" + str(state[key]).lower() + "</fluent-value>"
            msg = msg + "</observed-fluent>"
        msg = msg + "</turn>"
        return msg

    def build_round_end_msg(self, rew, round_reward):
        msg = "<round-end>"
        msg = msg + "<instance-name>" + self.problem + "</instance-name>"
        msg = msg + "<client-name>" + self.client + "</client-name>"
        msg = msg + "<round-num>" + str(self.currentround) + "</round-num>"
        msg = msg + "<round-reward>" +  str(round_reward) + "</round-reward>"
        msg = msg + "<turns-used>" + str(self.env.horizon) + "</turns-used>"
        msg = msg + "<time-left>1000</time-left>"
        msg = msg + "<immediate-reward>" + str(rew) + "</immediate-reward>"
        msg = msg + "</round-end>"
        return msg

    def build_session_end_msg(self):
        msg = "<session-end>"
        msg = msg + "<instance-name>" + self.problem + "</instance-name>"
        msg = msg + "<total-reward>" + str(self.total_reward) + "</total-reward>"
        msg = msg + "<rounds-used>" + str(self.currentround) + "</rounds-used>"
        msg = msg + "<time-used>0</time-used>"
        msg = msg + "<client-name>" + self.client + "</client-name>"
        msg = msg + "<session-id>0</session-id>"
        msg = msg + "<time-left>1000</time-left>"
        msg = msg + "</session-end>"
        return msg

    def process_init_session_request(self, data):
        parser = xmltree.XMLParser()
        root = xmltree.fromstring(data, parser)
        if (root.tag != "session-request"):
            print("Malformed session request message: session-request tag missing", flush=True)
            exit(1)
        self.problem = root.find("problem-name").text
        self.client = root.find("client-name").text
        input_language = root.find("input-language").text
        if (input_language != "rddl"):
            print("Malformed session request message: input language must be rddl", flush=True)
            exit(1)

    def process_round_request(self, data):
        parser = xmltree.XMLParser()
        root = xmltree.fromstring(data, parser)
        if (root.tag != "round-request"):
            print("Malformed round request message: round-request tag missing", flush=True)
            exit(1)
        execute = root.find("execute-policy").text.strip()
        if execute != "yes":
            print("Malformed round request message: policy must be executed", flush=True)
            exit(1)
        self.currentround += 1
        self.roundsleft -= 1

    def process_action(self, data):
        parser = xmltree.XMLParser()
        root = xmltree.fromstring(data, parser)
        if (root.tag != "actions"):
            print("Malformed action message: actions tag missing", flush=True)
            exit(1)
        actions = root.findall("action")
        result = {}
        for act in actions:
            name = act.find("action-name").text
            args = act.findall("action-arg")
            separator = PlanningModel.FLUENT_SEP
            for arg in args:
                name = name + separator + arg.text
                separator = PlanningModel.OBJECT_SEP
            value = act.find("action-value").text
            result[name] = value
        return result



from copy import deepcopy
import math
import numpy as np
from typing import Dict, List, Tuple, TYPE_CHECKING
import warnings

import gurobipy
from gurobipy import GRB

from pyRDDLGym.Core.ErrorHandling.RDDLException import print_stack_trace
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLTypeError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLNotImplementedError
from pyRDDLGym.Core.ErrorHandling.RDDLException import RDDLUndefinedVariableError

from pyRDDLGym.Core.Compiler.RDDLLevelAnalysis import RDDLLevelAnalysis
from pyRDDLGym.Core.Compiler.RDDLLiftedModel import RDDLLiftedModel
from pyRDDLGym.Core.Compiler.RDDLObjectsTracer import RDDLObjectsTracer
from pyRDDLGym.Core.Compiler.RDDLValueInitializer import RDDLValueInitializer
from pyRDDLGym.Core.Debug.Logger import Logger
from pyRDDLGym.Core.Grounder.RDDLGrounder import RDDLGrounder
from pyRDDLGym.Core.Simulator.RDDLSimulator import lngamma

if TYPE_CHECKING:
    from pyRDDLGym.Core.Gurobi.GurobiRDDLPlanner import GurobiRDDLPlan


class GurobiRDDLCompiler:
    '''Compile RDDL domains as Gurobi optimization problems.'''
    
    def __init__(self, rddl: RDDLLiftedModel,
                 plan: 'GurobiRDDLPlan',
                 allow_synchronous_state: bool=True,
                 rollout_horizon: int=None,
                 epsilon: float=1e-5,
                 float_range: Tuple[float, float]=(1e-15, 1e15),
                 model_params: Dict[str, object]={'NonConvex': 2},
                 piecewise_options: str='',
                 logger: Logger=None,
                 verbose: int=1) -> None:
        '''Creates a new compiler for formulating RDDL domains + instance as 
        a Gurobi mixed-integer non-linear optimization problem. In this base
        implemenation, a fixed subset of random variables are handled by
        determinization.
        
        :param rddl: the RDDL model
        :param plan: the plan or policy to optimize
        :param allow_synchronous_state: whether state-fluent can be synchronous
        :param rollout_horizon: length of the planning horizon (uses the RDDL
        defined horizon if None)
        :param epsilon: small positive constant used for comparing equality of
        real numbers in Gurobi constraints
        :param float_range: range of floating values that can be passed to 
        Gurobi to initialize fluents and non-fluents (values outside this range
        are clipped)
        :param model_params: dictionary of parameter name and values to
        pass to Gurobi model after compilation
        :param piecewise_options: a string of parameters to pass to Gurobi
        "options" parameter when creating constraints that contain piecewise
        linear approximations (e.g. cos, log, exp)
        :param logger: to log information about compilation to file
        :param verbose: whether to print nothing (0), summary (1),
        or detailed (2) messages to console
        '''
        self.plan = plan
        if rollout_horizon is None:
            rollout_horizon = rddl.horizon
        self.horizon = rollout_horizon
        self.discount = rddl.discount
        self.logger = logger
        self.verbose = int(verbose)
        
        # Gurobi-specific parameters
        self.epsilon = epsilon
        self.float_range = float_range
        self.model_params = model_params
        self.pw_options = piecewise_options
        
        # type conversion to Gurobi
        self.GUROBI_TYPES = {
            'int': GRB.INTEGER,
            'real': GRB.CONTINUOUS,
            'bool': GRB.BINARY
        }
        
        # ground out the domain
        grounder = RDDLGrounder(deepcopy(rddl._AST))
        self.rddl = grounder.Ground()
        
        # compile initial values
        if self.logger is not None:
            self.logger.clear()
        initializer = RDDLValueInitializer(self.rddl, logger=self.logger)
        self.init_values = initializer.initialize()
        
        # compute dependency graph for CPFs and sort them by evaluation order
        sorter = RDDLLevelAnalysis(
            self.rddl, allow_synchronous_state, logger=self.logger)
        self.levels = sorter.compute_levels()     
        
        # trace expressions to cache information to be used later
        tracer = RDDLObjectsTracer(self.rddl, logger=self.logger)
        self.traced = tracer.trace()
        
        # calculate no-op actions
        self.noop_actions = {}
        for (var, values) in self.init_values.items():
            if self.rddl.variable_types[var] == 'action-fluent':
                self.noop_actions[var] = values
    
    def summarize_hyperparameters(self):
        print(f'Gurobi compiler hyper-params:\n'
              f'    float_range       ={self.float_range}\n'
              f'    float_equality_tol={self.epsilon}\n'
              f'    lookahead_horizon ={self.horizon}\n'
              f'Gurobi model hyper-params:\n'
              f'    user_args         ={self.model_params}\n'
              f'    user_args_pw      ={self.pw_options}')
        
    # ===========================================================================
    # main compilation subroutines
    # ===========================================================================
        
    @staticmethod
    def get_variable_info(model: gurobipy.Model) -> Dict[str, Tuple]:
        '''Returns a dictionary summarizing variables in a given Gurobi Model, 
        where the keys are variable names, and values are tuples consisting of
        the variable data type and lower and upper bounds.
        '''
        model.update()
        result = {}
        for var in model.getVars():
            result[var.VarName] = (var.VType, var.LB, var.UB)
        return result
        
    def compile(self, init_values: Dict[str, object]=None,
                env: gurobipy.Env=None) -> \
        Tuple[gurobipy.Model, List[Dict[str, object]]]:
        '''Compiles and returns the current RDDL domain as a Gurobi optimization
        problem. Also returns action variables constructed during compilation 
        and policy parameter variables.
        
        :param init_values: override the initial values of fluents and
        non-fluents as defined in the RDDL file (if None, then the original
        values defined in the RDDL domain + instance are used instead)
        '''
        if self.verbose >= 1:
            self.summarize_hyperparameters()
            self.plan.summarize_hyperparameters()
            
        model = self._create_model(env=env)
        subs = self._compile_init_subs(init_values)
        
        params = self.plan.params(self, model) 
        objective, action_vars, _ = self._compile_rollout(
            model, self.plan, params, subs)
        model.setObjective(objective, GRB.MAXIMIZE)
        
        return model, action_vars, params
    
    def _compile_rollout(self, model, plan, params, subs, value_bounds: bool=False):
        
        # logging
        if self.logger is not None:
            message = f'[info] compiling initial bound information for Gurobi:'
            for (name, (_, _, lb, ub, _)) in subs.items():
                message += f'\n\t{name}, bounds=({lb}, {ub})'
            self.logger.log(message + '\n')  
                
        objective = 0
        all_action_vars = []
        all_next_state_vars = []
        for step in range(self.horizon):
            
            # add action fluent variables to model
            action_vars = plan.actions(self, model, params, step, subs)
            all_action_vars.append(action_vars)
            subs.update(action_vars)
            
            # add action constraints
            self._compile_maxnondef_constraint(model, subs)
            self._compile_action_preconditions(model, subs)
            
            # add constraint on state for the first step
            self._compile_state_invariants(model, subs)
                
            # evaluate CPFs and reward
            self._compile_cpfs(model, subs)
            reward, (lbr, ubr) = self._compile_reward(model, subs)
            discount = self.discount ** step
            objective += reward * discount
            
            # update state
            all_next_state_vars.append({})
            for (state, next_state) in self.rddl.next_state.items():
                subs[state] = subs[next_state]
                all_next_state_vars[-1][next_state] = subs[next_state]
            
            # update bounds on reward
            if value_bounds:
                if step == 0:
                    lb, ub = lbr, ubr
                else:
                    lb += lbr * discount
                    ub += ubr * discount
            
            # logging
            if self.logger is not None:
                message = f'[info] compiling bound information for Gurobi at epoch {step}:'
                for (name, (_, _, lb, ub, _)) in subs.items():
                    message += f'\n\t{name}, bounds=({lb}, {ub})'
                message += f'\n\treward, bounds=({lbr}, {ubr})'
                if value_bounds:
                    message += f'\n\tvalue_fn, bounds=({lb}, {ub})'
                self.logger.log(message + '\n')  
            
        if value_bounds:
            return objective, all_action_vars, all_next_state_vars, (lb, ub)
        else:
            return objective, all_action_vars, all_next_state_vars
    
    def _create_model(self, env: gurobipy.Env=None) -> gurobipy.Model:
        if env is None:
            warnings.warn(
                'Gurobi model created in default environment, not recommended.', 
                stacklevel=2)
            model = gurobipy.Model()
        else:
            model = gurobipy.Model(env=env)
        for (name, value) in self.model_params.items():
            model.setParam(name, value)
        return model 

    def _compile_init_subs(self, init_values=None) -> Dict[str, object]:
        if init_values is None:
            init_values = self.init_values
        rddl = self.rddl
        smallest, largest = self.float_range
        subs = {}
        for (var, value) in init_values.items():
            prange = rddl.variable_ranges[var]
            vtype = self.GUROBI_TYPES[prange]
            safe_value = value
            if rddl.variable_ranges[var] == 'real':
                if 0 < value < smallest:
                    safe_value = smallest
                elif -smallest < value < 0:
                    safe_value = -smallest
                elif value > largest:
                    safe_value = largest
                elif value < -largest:
                    safe_value = -largest
            lb, ub = GurobiRDDLCompiler._fix_bounds(safe_value, safe_value)
            subs[var] = (safe_value, vtype, lb, ub, False)
        return subs
        
    def _compile_action_preconditions(self, model, subs) -> None:
        for precondition in self.rddl.preconditions:
            indicator, *_, symb = self._gurobi(precondition, model, subs)
            if symb:
                model.addConstr(indicator == 1)
    
    def _compile_state_invariants(self, model, subs) -> None:
        for invariant in self.rddl.invariants:
            indicator, *_, symb = self._gurobi(invariant, model, subs)
            if symb:
                model.addConstr(indicator == 1)
        
    def _compile_maxnondef_constraint(self, model, subs) -> None:
        rddl = self.rddl
        num_bool, sum_bool = 0, 0
        for (action, prange) in rddl.actionsranges.items():
            if prange == 'bool':
                var, *_ = subs[action]
                num_bool += 1
                sum_bool += var
        if rddl.max_allowed_actions < num_bool:
            model.addConstr(sum_bool <= rddl.max_allowed_actions)
            
    def _compile_cpfs(self, model, subs) -> None:
        rddl = self.rddl
        for cpfs in self.levels.values():
            for cpf in sorted(cpfs):
                _, expr = rddl.cpfs[cpf]
                subs[cpf] = self._gurobi(expr, model, subs)
    
    def _compile_reward(self, model, subs) -> object:
        reward, _, lb, ub, _ = self._gurobi(self.rddl.reward, model, subs)
        return reward, (lb, ub)
    
    # ===========================================================================
    # start of compilation subroutines
    # ===========================================================================
    
    # IMPORTANT: all helper methods below must return either a Gurobi variable
    # or a constant as the first argument
    def _gurobi(self, expr, model, subs):
        etype, _ = expr.etype
        if etype == 'constant':
            return self._gurobi_constant(expr, model, subs)
        elif etype == 'pvar':
            return self._gurobi_pvar(expr, model, subs)
        elif etype == 'arithmetic':
            return self._gurobi_arithmetic(expr, model, subs)
        elif etype == 'relational':
            return self._gurobi_relational(expr, model, subs)
        elif etype == 'boolean':
            return self._gurobi_logical(expr, model, subs)
        elif etype == 'func':
            return self._gurobi_function(expr, model, subs)
        elif etype == 'control':
            return self._gurobi_control(expr, model, subs)
        elif etype == 'randomvar':
            return self._gurobi_random(expr, model, subs)
        else:
            raise RDDLNotImplementedError(
                f'Expression type {etype} is not supported in Gurobi compiler.\n' + 
                print_stack_trace(expr))
            
    def _add_var(self, model, vtype, lb=-GRB.INFINITY, ub=GRB.INFINITY, name=''):
        '''Add a generic variable to the Gurobi model.'''
        return model.addVar(vtype=vtype, lb=lb, ub=ub, name=name)
    
    def _add_bool_var(self, model, name=''):
        '''Add a BINARY variable to the Gurobi model.'''
        return self._add_var(model, GRB.BINARY, 0, 1, name=name)
    
    def _add_real_var(self, model, lb=-GRB.INFINITY, ub=GRB.INFINITY, name=''):
        '''Add a CONTINUOUS variable to the Gurobi model.'''
        return self._add_var(model, GRB.CONTINUOUS, lb, ub, name=name)
    
    def _add_int_var(self, model, lb=-GRB.INFINITY, ub=GRB.INFINITY, name=''):
        '''Add a INTEGER variable to the Gurobi model.'''
        return self._add_var(model, GRB.INTEGER, lb, ub, name=name)
    
    # ===========================================================================
    # leaves
    # ===========================================================================
    
    @staticmethod
    def _fix_bounds(lb, ub):
        assert (not math.isnan(lb))
        assert (not math.isnan(ub))
        assert (ub >= lb)
        lb = max(min(lb, GRB.INFINITY), -GRB.INFINITY)
        ub = max(min(ub, GRB.INFINITY), -GRB.INFINITY)
        return lb, ub
    
    @staticmethod
    def _fix_bounds_abs(lb, ub):
        if lb >= 0:
            pass
        elif ub <= 0:
            lb, ub = -ub, -lb
        else:
            lb, ub = 0, max(abs(lb), abs(ub))
        return GurobiRDDLCompiler._fix_bounds(lb, ub)
    
    @staticmethod
    def _fix_product_underflow(a, b):
        a1, b1 = abs(a), abs(b)
        if (a1 <= 0 or b1 <= 0) or (a1 >= 1 or b1 >= 1):
            return a * b
        tiny = np.finfo(np.float64).tiny
        if a1 < tiny / b1:
            warnings.warn(
                f'Underflow prevented for {a} * {b} by setting to zero.', 
                stacklevel=2)
            return 0.0
        return a * b
        
    @staticmethod
    def _fix_bounds_prod(lb1, ub1, lb2, ub2):
        lbub = (GurobiRDDLCompiler._fix_product_underflow(lb1, lb2), 
                GurobiRDDLCompiler._fix_product_underflow(lb1, ub2), 
                GurobiRDDLCompiler._fix_product_underflow(ub1, lb2), 
                GurobiRDDLCompiler._fix_product_underflow(ub1, ub2))
        return GurobiRDDLCompiler._fix_bounds(min(lbub), max(lbub))
        
    def _gurobi_constant(self, expr, model, subs):
        
        # get the cached value of this constant
        value = self.traced.cached_sim_info(expr)
        
        # infer type of value and assign to Gurobi type
        if isinstance(value, bool):
            vtype = GRB.BINARY
        elif isinstance(value, int):
            vtype = GRB.INTEGER
        elif isinstance(value, float):
            vtype = GRB.CONTINUOUS
        else:
            raise RDDLNotImplementedError(
                f'Range of {value} is not supported in Gurobi compiler.')
        
        # bounds form a singleton set containing the cached value
        lb, ub = GurobiRDDLCompiler._fix_bounds(value, value)
        return lb, vtype, lb, ub, False

    def _gurobi_pvar(self, expr, model, subs):
        var, _ = expr.args
        
        # domain object converted to canonical index
        is_value, value = self.traced.cached_sim_info(expr)
        if is_value:
            lb, ub = GurobiRDDLCompiler._fix_bounds(value, value)
            return lb, GRB.INTEGER, lb, ub, False
        
        # extract variable value
        value = subs.get(var, None)
        if value is None:
            raise RDDLUndefinedVariableError(
                f'Variable <{var}> is referenced before assignment.\n' + 
                print_stack_trace(expr))
        return value
    
    # ===========================================================================
    # arithmetic
    # ===========================================================================
    
    @staticmethod
    def _promote_vtype(vtype1, vtype2):
        if vtype1 == GRB.BINARY:
            return vtype2
        elif vtype2 == GRB.BINARY:
            return vtype1
        elif vtype1 == GRB.INTEGER:
            return vtype2
        elif vtype2 == GRB.INTEGER:
            return vtype1
        else:
            assert (vtype1 == vtype2 == GRB.CONTINUOUS)
            return vtype1
    
    @staticmethod
    def _at_least_int(vtype):
        return GurobiRDDLCompiler._promote_vtype(vtype, GRB.INTEGER)
    
    def _gurobi_arithmetic(self, expr, model, subs):
        _, op = expr.etype
        args = expr.args        
        n = len(args)
        
        # unary negation
        if n == 1 and op == '-':
            arg, = args
            gterm, vtype, lb, ub, symb = self._gurobi(arg, model, subs)
            vtype = GurobiRDDLCompiler._at_least_int(vtype)
            lb, ub = GurobiRDDLCompiler._fix_bounds(-1 * ub, -1 * lb)
            negexpr = -1 * gterm
            
            # assign negative to a new variable
            if symb: 
                res = self._add_var(model, vtype, lb, ub)
                model.addConstr(res == negexpr)
            else:
                res = lb = ub = negexpr           
            return res, vtype, lb, ub, symb
        
        # binary operations
        elif n >= 1:
            results = [self._gurobi(arg, model, subs) for arg in args]
            
            # unwrap addition to binary operations
            if op == '+':
                sumexpr, vtype, lb, ub, symb = results[0]
                vtype = GurobiRDDLCompiler._at_least_int(vtype)
                res = sumexpr
                for (gterm2, vtype2, lb2, ub2, symb2) in results[1:]:
                    sumexpr = sumexpr + gterm2
                    vtype = GurobiRDDLCompiler._promote_vtype(vtype, vtype2)
                    lb, ub = GurobiRDDLCompiler._fix_bounds(lb + lb2, ub + ub2)
                    symb = symb or symb2
                
                    # assign sum to a new variable
                    if symb:
                        res = self._add_var(model, vtype, lb, ub)
                        model.addConstr(res == sumexpr)
                        sumexpr = res
                    else:
                        res = lb = ub = sumexpr                   
                return res, vtype, lb, ub, symb
            
            # unwrap multiplication to binary operations
            elif op == '*':
                prodexpr, vtype, lb, ub, symb = results[0]
                vtype = GurobiRDDLCompiler._at_least_int(vtype)
                res = prodexpr
                for (gterm2, vtype2, lb2, ub2, symb2) in results[1:]:
                    prodexpr = prodexpr * gterm2
                    vtype = GurobiRDDLCompiler._promote_vtype(vtype, vtype2)
                    lb, ub = GurobiRDDLCompiler._fix_bounds_prod(lb, ub, lb2, ub2)
                    symb = symb or symb2
                    
                    # assign product to a new variable
                    if symb: 
                        res = self._add_var(model, vtype, lb, ub)
                        model.addConstr(res == prodexpr)
                        prodexpr = res
                    else:
                        res = lb = ub = prodexpr                    
                return res, vtype, lb, ub, symb
            
            # subtraction
            elif n == 2 and op == '-':
                gterm1, vtype1, lb1, ub1, symb1 = results[0]
                gterm2, vtype2, lb2, ub2, symb2 = results[1]
                vtype = GurobiRDDLCompiler._promote_vtype(vtype1, vtype2)
                vtype = GurobiRDDLCompiler._at_least_int(vtype)
                diffexpr = gterm1 - gterm2
                symb = symb1 or symb2
                
                # assign difference to a new variable
                if symb:
                    lb, ub = GurobiRDDLCompiler._fix_bounds(lb1 - ub2, ub1 - lb2)
                    res = self._add_var(model, vtype, lb, ub)
                    model.addConstr(res == diffexpr)
                else:
                    res = lb = ub = diffexpr                
                return res, vtype, lb, ub, symb
            
            # implement z = x / y as a constraint z * y = x
            elif n == 2 and op == '/': 
                gterm1, _, lb1, ub1, symb1 = results[0]
                gterm2, _, lb2, ub2, symb2 = results[1]
                symb = symb1 or symb2
                
                if symb:
                    if symb2: 
                        if 0 > lb2 and 0 < ub2:
                            lb2, ub2 = -GRB.INFINITY, GRB.INFINITY
                        elif lb2 == 0 and ub2 == 0:
                            lb2, ub2 = GRB.INFINITY, GRB.INFINITY
                        elif lb2 == 0:
                            lb2, ub2 = 1 / ub2, GRB.INFINITY
                        elif ub2 == 0:
                            lb2, ub2 = -GRB.INFINITY, 1 / lb2
                        else:
                            lb2, ub2 = 1 / ub2, 1 / lb2
                    else:
                        lb2 = ub2 = 1 / gterm2
                    lb, ub = GurobiRDDLCompiler._fix_bounds_prod(lb1, ub1, lb2, ub2)      
                                  
                    res = self._add_real_var(model, lb, ub)
                    model.addConstr(res * gterm2 == gterm1)    
                else:
                    res = lb = ub = gterm1 / gterm2    
                return res, GRB.CONTINUOUS, lb, ub, symb
        
        raise RDDLNotImplementedError(
            f'Arithmetic operator {op} with {n} arguments is not '
            f'supported in Gurobi compiler.\n' + 
            print_stack_trace(expr))
    
    # ===========================================================================
    # boolean
    # ===========================================================================
    
    def _gurobi_relational(self, expr, model, subs):
        _, op = expr.etype
        args = expr.args        
        n = len(args)
        
        if n == 2:
            lhs, rhs = args
            glhs, vtype1, lb1, ub1, symb1 = self._gurobi(lhs, model, subs)
            grhs, vtype2, lb2, ub2, symb2 = self._gurobi(rhs, model, subs)
            vtype = GurobiRDDLCompiler._promote_vtype(vtype1, vtype2)
            vtype = GurobiRDDLCompiler._at_least_int(vtype)
            symb = symb1 or symb2
            
            # convert <= to >=, < to >, etc.
            if op == '<=' or op == '<':
                glhs, grhs = grhs, glhs
                op = '>=' if op == '<=' else '>'
            diffexpr = glhs - grhs
            
            # assign comparison operator to binary variable
            if op == '==': 
                if symb:
                    diff_var = self._add_var(model, vtype, lb1 - ub2, ub1 - lb2)
                    model.addConstr(diff_var == diffexpr)
                    
                    lb, ub = GurobiRDDLCompiler._fix_bounds_abs(lb1 - ub2, ub1 - lb2)
                    abs_diff = self._add_var(model, vtype, lb, ub)
                    model.addGenConstrAbs(abs_diff, diff_var)
                                 
                    res = self._add_bool_var(model)
                    model.addConstr((res == 1) >> (abs_diff <= 0))
                    model.addConstr((res == 0) >> (abs_diff >= self.epsilon))
                    lb, ub = 0, 1
                else:
                    res = bool(glhs == grhs)
                    lb = ub = int(res)
                return res, GRB.BINARY, lb, ub, symb
            
            elif op == '>=':
                if symb:
                    res = self._add_bool_var(model)
                    model.addConstr((res == 1) >> (diffexpr >= 0))
                    model.addConstr((res == 0) >> (diffexpr <= -self.epsilon))
                    lb, ub = 0, 1
                else:
                    res = bool(glhs >= grhs)
                    lb = ub = int(res)
                return res, GRB.BINARY, lb, ub, symb
            
            elif op == '~=':
                if symb:
                    diff_var = self._add_var(model, vtype, lb1 - ub2, ub1 - lb2)
                    model.addConstr(diff_var == diffexpr)
                    
                    lb, ub = GurobiRDDLCompiler._fix_bounds_abs(lb1 - ub2, ub1 - lb2)
                    abs_diff = self._add_var(model, vtype, lb, ub)
                    model.addGenConstrAbs(abs_diff, diff_var)
                    
                    res = self._add_bool_var(model)
                    model.addConstr((res == 1) >> (abs_diff >= self.epsilon))
                    model.addConstr((res == 0) >> (abs_diff <= 0))
                    lb, ub = 0, 1
                else: 
                    res = bool(glhs != grhs)
                    lb = ub = int(res)
                return res, GRB.BINARY, lb, ub, symb
            
            elif op == '>':
                if symb:
                    res = self._add_bool_var(model)
                    model.addConstr((res == 1) >> (diffexpr >= self.epsilon))
                    model.addConstr((res == 0) >> (diffexpr <= 0))
                    lb, ub = 0, 1
                else:
                    res = bool(glhs > grhs)
                    lb = ub = int(res)
                return res, GRB.BINARY, lb, ub, symb
            
        raise RDDLNotImplementedError(
            f'Relational operator {op} with {n} arguments is not '
            f'supported in Gurobi compiler.\n' + 
            print_stack_trace(expr))
    
    def _gurobi_logical(self, expr, model, subs):
        _, op = expr.etype
        if op == '&':
            op = '^'
        args = expr.args        
        n = len(args)
        
        # unary negation ~z of z is a variable y such that y + z = 1
        if n == 1 and op == '~':
            arg, = args
            gterm, *_, symb = self._gurobi(arg, model, subs)
            if symb:
                res = self._add_bool_var(model)
                model.addConstr(res + gterm == 1)
                lb, ub = 0, 1
            else:
                if not isinstance(gterm, (bool, np.bool_)):
                    raise RDDLTypeError(
                        f'Constant expression is of type '
                        f'{type(gterm)}, expected bool or np.bool_' + 
                        '\n' + print_stack_trace(arg))
                res = not bool(gterm)
                lb = ub = int(res)            
            return res, GRB.BINARY, lb, ub, symb
            
        # binary operations
        elif n >= 1:
            results = [self._gurobi(arg, model, subs) for arg in args]
            gterms = [result[0] for result in results]
            symbs = [result[-1] for result in results]
            symb = any(symbs)
            
            # any non-variables must be converted to variables
            if symb:
                for (i, gterm) in enumerate(gterms):
                    if not symbs[i]:
                        if not isinstance(gterm, (bool, np.bool_)):
                            raise RDDLTypeError(
                                f'Constant expression is of type '
                                f'{type(gterm)}, expected bool or np.bool_' + 
                                '\n' + print_stack_trace(args[i]))
                        var = self._add_bool_var(model)
                        model.addConstr(var == bool(gterm))
                        gterms[i] = var
                        symbs[i] = True
            
            # unwrap AND to binary operations
            if op == '^':
                if symb:
                    res = self._add_bool_var(model)
                    model.addGenConstrAnd(res, gterms)
                    lb, ub = 0, 1
                else:
                    res = all(gterms)   
                    lb = ub = int(res)                 
                return res, GRB.BINARY, lb, ub, symb
            
            # unwrap OR to binary operations
            elif op == '|':
                if symb:
                    res = self._add_bool_var(model)
                    model.addGenConstrOr(res, gterms)
                    lb, ub = 0, 1
                else:
                    res = any(gterms)    
                    lb = ub = int(res)                
                return res, GRB.BINARY, lb, ub, symb
            
            # unwrap => to binary operations
            elif op == '=>' and n == 2:
                gterm1, gterm2 = gterms
                if symb:
                    not1 = self._add_bool_var(model)
                    model.addConstr(not1 + gterm1 == 1)
                              
                    res = self._add_bool_var(model)
                    model.addGenConstrOr(res, [not1, gterm2])
                    lb, ub = 0, 1             
                else:
                    res = (gterm1 <= gterm2)
                    lb = ub = int(res)
                return res, GRB.BINARY, lb, ub, symb                
                            
        raise RDDLNotImplementedError(
            f'Logical operator {op} with {n} arguments is not '
            f'supported in Gurobi compiler.\n' + 
            print_stack_trace(expr))
    
    # ===========================================================================
    # function
    # ===========================================================================

    @staticmethod
    def _log(x):
        if x <= 0:
            return -GRB.INFINITY
        else:
            return math.log(x)
    
    def _gurobi_positive(self, model, gterm, vtype, lb, ub):
        lb, ub = max(lb, 0), max(ub, 0)
        res = self._add_var(model, vtype, lb, ub)
        model.addGenConstrMax(res, [gterm], constant=0)
        return res, lb, ub
                    
    def _gurobi_function(self, expr, model, subs):
        _, name = expr.etype
        args = expr.args
        n = len(args)
        
        # unary functions
        if n == 1:
            arg, = args
            gterm, vtype, lb, ub, symb = self._gurobi(arg, model, subs)
            vtype = GurobiRDDLCompiler._at_least_int(vtype)
            
            if name == 'abs': 
                if symb:
                    lb, ub = GurobiRDDLCompiler._fix_bounds_abs(lb, ub)                    
                    res = self._add_var(model, vtype, lb, ub)
                    model.addGenConstrAbs(res, gterm)                    
                else:
                    res = lb = ub = abs(gterm)           
                return res, vtype, lb, ub, symb
            
            elif name == 'sgn':
                if symb:
                    if lb > 0:
                        res = lb = ub = 1
                        symb = False
                    elif ub < 0:
                        res = lb = ub = -1
                        symb = False
                    else:
                        pos = self._add_bool_var(model)
                        model.addConstr((pos == 1) >> (gterm >= self.epsilon))
                        model.addConstr((pos == 0) >> (gterm <= 0))
                        neg = self._add_bool_var(model)
                        model.addConstr((neg == 1) >> (gterm <= -self.epsilon))
                        model.addConstr((neg == 0) >> (gterm >= 0))
                        res = self._add_int_var(model, lb=-1, ub=1)
                        model.addConstr(res + neg == pos)
                        lb, ub = -1, 1
                else:
                    if gterm > 0:
                        res = lb = ub = 1
                    elif gterm < 0:
                        res = lb = ub = -1
                    else:
                        res = lb = ub = 0
                return res, GRB.INTEGER, lb, ub, symb
                
            elif name == 'floor':
                if symb:
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        math.floor(lb), math.floor(ub))
                    res = self._add_int_var(model, lb, ub)
                    model.addConstr(res <= gterm)
                    model.addConstr(res + 1 >= gterm + self.epsilon)                    
                else:
                    res = lb = ub = int(math.floor(gterm))
                return res, GRB.INTEGER, lb, ub, symb
            
            elif name == 'ceil':
                if symb:
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        math.ceil(lb), math.ceil(ub))
                    res = self._add_int_var(model, lb, ub)
                    model.addConstr(res >= gterm)
                    model.addConstr(res - 1 <= gterm - self.epsilon)
                else:
                    res = lb = ub = int(math.ceil(gterm))
                return res, GRB.INTEGER, lb, ub, symb
                
            elif name == 'cos':
                if symb:
                    lb, ub = -1.0, 1.0
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrCos(gterm, res, options=self.pw_options)
                else:
                    res = lb = ub = math.cos(gterm)      
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'sin':
                if symb:
                    lb, ub = -1.0, 1.0
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrSin(gterm, res, options=self.pw_options)
                else:
                    res = lb = ub = math.sin(gterm)      
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'tan':
                if symb:
                    lb, ub = -GRB.INFINITY, GRB.INFINITY
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrTan(gterm, res, options=self.pw_options)
                else:
                    res = lb = ub = math.tan(gterm)      
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'exp':
                if symb: 
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        math.exp(lb), math.exp(ub))
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrExp(gterm, res, options=self.pw_options)
                else:
                    res = lb = ub = math.exp(gterm)      
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'ln': 
                if symb:
                    arg, lb, ub = self._gurobi_positive(model, gterm, vtype, lb, ub)                    
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        GurobiRDDLCompiler._log(lb), GurobiRDDLCompiler._log(ub))
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrLog(arg, res, options=self.pw_options)
                else:
                    res = lb = ub = math.log(gterm)      
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'sqrt':
                if symb: 
                    arg, lb, ub = self._gurobi_positive(model, gterm, vtype, lb, ub)                    
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        math.sqrt(lb), math.sqrt(ub))
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrPow(arg, res, 0.5, options=self.pw_options)
                else:
                    res = lb = ub = math.sqrt(gterm)     
                return res, GRB.CONTINUOUS, lb, ub, symb
        
        # binary functions
        elif n == 2:
            arg1, arg2 = args
            gterm1, vtype1, lb1, ub1, symb1 = self._gurobi(arg1, model, subs)
            gterm2, vtype2, lb2, ub2, symb2 = self._gurobi(arg2, model, subs)
            vtype = GurobiRDDLCompiler._promote_vtype(vtype1, vtype2)
            vtype = GurobiRDDLCompiler._at_least_int(vtype)
            symb = symb1 or symb2
            
            if name == 'min': 
                if symb: 
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        min(lb1, lb2), min(ub1, ub2))
                    res = self._add_var(model, vtype, lb, ub)
                    model.addGenConstrMin(res, [gterm1, gterm2])
                else:
                    res = lb = ub = min(gterm1, gterm2)  
                return res, vtype, lb, ub, symb
            
            elif name == 'max':
                if symb:
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        max(lb1, lb2), max(ub1, ub2))
                    res = self._add_var(model, vtype, lb, ub)
                    model.addGenConstrMax(res, [gterm1, gterm2])
                else:
                    res = lb = ub = max(gterm1, gterm2)  
                return res, vtype, lb, ub, symb
            
            elif name == 'pow':
                if symb: 
                    # argument must be non-negative
                    base, lb1, ub1 = self._gurobi_positive(
                        model, gterm1, vtype1, lb1, ub1)   
                                        
                    # compute bounds on pow
                    loglb = GurobiRDDLCompiler._log(lb1)
                    logub = GurobiRDDLCompiler._log(ub1)                    
                    loglu = (loglb * lb2, loglb * ub2, logub * lb2, logub * ub2)
                    lb, ub = GurobiRDDLCompiler._fix_bounds(
                        math.exp(min(loglu)), math.exp(max(loglu)))
                    
                    # assign pow to new variable
                    res = self._add_real_var(model, lb, ub)
                    model.addGenConstrPow(
                        base, res, gterm2, options=self.pw_options)
                else:
                    res = lb = ub = math.pow(gterm1, gterm2)         
                return res, GRB.CONTINUOUS, lb, ub, symb
            
            elif name == 'mod':
                if symb:
                    # second argument must be non-negative
                    gterm2, lb2, ub2 = self._gurobi_positive(
                        model, gterm2, vtype2, lb2, ub2)  
                    
                    # compute r = x % y as x = y * q + r where 0 <= r < y
                    lb, ub = 0, max(0, ub2 - 1)
                    res = self._add_int_var(model, lb, ub)
                    quotient = self._add_int_var(model)
                    model.addConstr(gterm1 == gterm2 * quotient + res)                    
                else:
                    res = lb = ub = gterm1 % gterm2
                return res, GRB.INTEGER, lb, ub, symb
            
            elif name == 'fmod':
                if symb:
                    # second argument must be non-negative
                    gterm2, lb2, ub2 = self._gurobi_positive(
                        model, gterm2, vtype2, lb2, ub2)  
                    
                    # compute r = x % y as x = y * q + r where 0 <= r < y
                    lb, ub = 0, max(0, ub2 - self.epsilon)
                    res = self._add_real_var(model, lb, ub)
                    quotient = self._add_int_var(model)
                    model.addConstr(gterm1 == gterm2 * quotient + res)                    
                else:
                    res = lb = ub = gterm1 % gterm2
                return res, GRB.CONTINUOUS, lb, ub, symb
                
        raise RDDLNotImplementedError(
            f'Function operator {name} with {n} arguments is not '
            f'supported in Gurobi compiler.\n' + 
            print_stack_trace(expr))

    # ===========================================================================
    # control flow
    # ===========================================================================
    
    def _gurobi_control(self, expr, model, subs):
        _, op = expr.etype
        args = expr.args
        n = len(args)
        
        if n == 3 and op == 'if':
            pred, arg1, arg2 = args
            gpred, *_, symbp = self._gurobi(pred, model, subs)
            gterm1, vtype1, lb1, ub1, symb1 = self._gurobi(arg1, model, subs)
            gterm2, vtype2, lb2, ub2, symb2 = self._gurobi(arg2, model, subs)
            vtype = GurobiRDDLCompiler._promote_vtype(vtype1, vtype2)
            
            # assign if to new variable
            if symbp:
                lb, ub = GurobiRDDLCompiler._fix_bounds(min(lb1, lb2), max(ub1, ub2))
                res = self._add_var(model, vtype, lb, ub)
                model.addConstr((gpred == 1) >> (res == gterm1))
                model.addConstr((gpred == 0) >> (res == gterm2))
                symb = True
            else:
                if not isinstance(gpred, (bool, np.bool_)):
                    raise RDDLTypeError(
                        f'Constant expression is of type '
                        f'{type(gpred)}, expected bool or np.bool_' + 
                        '\n' + print_stack_trace(pred))
                if bool(gpred):
                    res, lb, ub, symb = gterm1, lb1, ub1, symb1
                else:
                    res, lb, ub, symb = gterm2, lb2, ub2, symb2
            return res, vtype, lb, ub, symb
            
        raise RDDLNotImplementedError(
            f'Control flow {op} with {n} arguments is not '
            f'supported in Gurobi compiler.\n' + 
            print_stack_trace(expr))

    # ===========================================================================
    # random variables
    # ===========================================================================
    
    def _gurobi_random(self, expr, model, subs):
        _, name = expr.etype
        if name == 'KronDelta':
            return self._gurobi_kron(expr, model, subs)
        elif name == 'DiracDelta':
            return self._gurobi_dirac(expr, model, subs)
        elif name == 'Uniform':
            return self._gurobi_uniform(expr, model, subs)
        elif name == 'Bernoulli':
            return self._gurobi_bernoulli(expr, model, subs)
        elif name == 'Normal':
            return self._gurobi_normal(expr, model, subs)
        elif name == 'Poisson':
            return self._gurobi_poisson(expr, model, subs)
        elif name == 'Exponential':
            return self._gurobi_exponential(expr, model, subs)
        elif name == 'Gamma':
            return self._gurobi_gamma(expr, model, subs)
        elif name == 'Weibull':
            return self._gurobi_weibull(expr, model, subs)
        else:
            raise RDDLNotImplementedError(
                f'Distribution {name} is not supported in Gurobi compiler.\n' + 
                print_stack_trace(expr))
    
    def _gurobi_kron(self, expr, model, subs): 
        arg, = expr.args
        return self._gurobi(arg, model, subs)
    
    def _gurobi_dirac(self, expr, model, subs):
        arg, = expr.args
        return self._gurobi(arg, model, subs)
    
    def _gurobi_uniform(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Uniform(a, b) --> (a + b) / 2.',  stacklevel=2)
            
        arg1, arg2 = expr.args
        gterm1, _, lb1, ub1, symb1 = self._gurobi(arg1, model, subs)
        gterm2, _, lb2, ub2, symb2 = self._gurobi(arg2, model, subs)
        
        # determinize uniform as (lower + upper) / 2        
        symb = symb1 or symb2
        midexpr = (gterm1 + gterm2) / 2
        if symb:
            lb, ub = GurobiRDDLCompiler._fix_bounds(
                (lb1 + lb2) / 2, (ub1 + ub2) / 2)
            res = self._add_real_var(model, lb, ub)
            model.addConstr(res == midexpr)            
        else:
            res = lb = ub = midexpr
        return res, GRB.CONTINUOUS, lb, ub, symb
        
    def _gurobi_bernoulli(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Bernoulli(p) --> p',  stacklevel=2)
            
        arg, = expr.args
        gterm, _, lb, ub, symb = self._gurobi(arg, model, subs)
        
        # determinize bernoulli as indicator of p > 0.5
        if symb:
            res = self._add_bool_var(model)
            model.addConstr((res == 1) >> (gterm >= 0.5 + self.epsilon))
            model.addConstr((res == 0) >> (gterm <= 0.5))
            lb, ub = 0, 1
        else:
            res = bool(gterm > 0.5)
            lb = ub = int(res)
        return res, GRB.BINARY, lb, ub, symb
        
    def _gurobi_normal(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Normal(m, v) --> m',  stacklevel=2)
            
        mean, _ = expr.args
        gterm, _, lb, ub, symb = self._gurobi(mean, model, subs)
        
        # determinize Normal as mean
        return gterm, GRB.CONTINUOUS, lb, ub, symb
    
    def _gurobi_poisson(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Poisson(l) --> l',  stacklevel=2)
            
        rate, = expr.args
        gterm, _, lb, ub, symb = self._gurobi(rate, model, subs)
        
        # determinize Poisson as rate
        return gterm, GRB.CONTINUOUS, lb, ub, symb
    
    def _gurobi_exponential(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Exponential(l) --> l',  stacklevel=2)
            
        scale, = expr.args
        gterm, _, lb, ub, symb = self._gurobi(scale, model, subs)
        
        # determinize Exponential as scale
        return gterm, GRB.CONTINUOUS, lb, ub, symb

    def _gurobi_gamma(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Gamma(r, s) --> r * s',  stacklevel=2)
            
        shape, scale = expr.args
        gterm1, _, lb1, ub1, symb1 = self._gurobi(shape, model, subs)
        gterm2, _, lb2, ub2, symb2 = self._gurobi(scale, model, subs)
        
        # determinize gamma as shape * scale
        prodexpr = gterm1 * gterm2
        symb = symb1 or symb2
        if symb:
            lb, ub = GurobiRDDLCompiler._fix_bounds_prod(lb1, ub1, lb2, ub2)
            res = self._add_real_var(model, lb, ub)
            model.addConstr(res == prodexpr)
        else:
            res = lb = ub = prodexpr     
        return res, GRB.CONTINUOUS, lb, ub, symb
    
    def _gurobi_weibull(self, expr, model, subs):
        if self.verbose >= 2:
            warnings.warn('Using the replacement rule: '
                          'Weibull(s, l) --> l * Gamma(1 + 1/s)',  stacklevel=2)
        
        shape, scale = expr.args
        gterm1, _, _, _, symb1 = self._gurobi(shape, model, subs)
        gterm2, _, lb2, ub2, symb2 = self._gurobi(scale, model, subs)
        
        # check shape is non symbolic
        if symb1:
            raise RDDLNotImplementedError(
                'Weibull symbolic expression for shape parameter is not '
                'supported in Gurobi compiler.\n' + 
                print_stack_trace(expr))
        
        # estimate mean as scale * Gamma(1 + 1/shape)
        gampart = np.exp(lngamma(1. + 1. / gterm1))
        gampart = max(min(gampart, GRB.INFINITY), -GRB.INFINITY)
        prodexpr = gterm2 * gampart
        symb = symb1 or symb2
        if symb:
            lb, ub = GurobiRDDLCompiler._fix_bounds_prod(lb2, ub2, gampart, gampart)
            res = self._add_real_var(model, lb, ub)
            model.addConstr(res == prodexpr)
        else:
            res = lb = ub = prodexpr
        return res, GRB.CONTINUOUS, lb, ub, symb
        


'''This example runs the Gurobi planner. 

The syntax is:

    python GurobiExample.py <domain> <instance> <horizon>
    
where:
    <domain> is the name of a domain located in the /Examples directory
    <instance> is the instance number
    <horizon> is a positive integer representing the lookahead horizon
'''
import sys

from pyRDDLGym.Core.Env.RDDLEnv import RDDLEnv
from pyRDDLGym.Core.Gurobi.GurobiRDDLPlanner import GurobiStraightLinePlan
from pyRDDLGym.Core.Gurobi.GurobiRDDLPlanner import GurobiOnlineController
from pyRDDLGym.Examples.ExampleManager import ExampleManager


def main(domain, instance, horizon):
    
    # create the environment
    info = ExampleManager.GetEnvInfo(domain)    
    env = RDDLEnv.build(info, instance, enforce_action_constraints=True)
    
    # create the controller
    controller = GurobiOnlineController(rddl=env.model,
                                        plan=GurobiStraightLinePlan(),
                                        rollout_horizon=horizon,
                                        model_params={'NonConvex': 2, 'OutputFlag': 1})
    controller.evaluate(env, verbose=True, render=True)
    
    env.close()

            
if __name__ == "__main__":
    args = sys.argv[1:]
    if len(args) < 3:
        print('python GurobiExample.py <domain> <instance> <horizon>')
        exit(1)
    domain, instance, horizon = args[:3]
    horizon = int(horizon)
    main(domain, instance, horizon)
    
